# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Packaging for execution on Cloud."""
from typing import Any, Optional

from xmanager import xm
from xmanager.cloud import auth
from xmanager.cloud import build_image
from xmanager.cloud import docker_lib
from xmanager.docker import docker_adapter
from xmanager.xm import pattern_matching
from xmanager.xm_local import executables as local_executables
from xmanager.xm_local import executors as local_executors


def _get_push_image_tag(executor_spec: xm.ExecutorSpec) -> Optional[str]:
  """Get the push_image_tag from executor or None."""

  def _get_push_image_tag_caip(spec: local_executors.CaipSpec):
    return spec.push_image_tag

  def _get_push_image_tag_kubernetes(spec: local_executors.KubernetesSpec):
    return spec.push_image_tag

  def _throw_on_unknown_executor(executor: Any):
    raise TypeError(f'Unsupported executor specification: {executor!r}. ')

  return pattern_matching.match(
      _get_push_image_tag_caip,
      _get_push_image_tag_kubernetes,
      _throw_on_unknown_executor,
  )(
      executor_spec)


def _package_container(packageable: xm.Packageable,
                       container: xm.Container) -> xm.Executable:
  """Matcher method for packaging `xm.Container`.

  If the user builds an image with the same repository name as an image in a
  remote Cloud registry, the user should push the image before running XManager,
  because XManager will not push a local image in the packaging step.

  Unless the container's image path already matches the GCR project prefix, we
  always pull the container and push it to push_image_tag location. This avoids
  a potential permissions error where the user has permissions to read an image,
  but the Cloud service agent does not have permissions. If container.image_path
  already points to the GCR project, we skip pushing because the image should
  already be in the destination location.

  Args:
    packageable: Packageable containing Executor and Executable.
    container: Container specifying image path.

  Returns:
    GoogleContainerRegistryImage Executable.
  """
  gcr_project_prefix = 'gcr.io/' + auth.get_project_name()
  if container.image_path.startswith(gcr_project_prefix):
    return local_executables.GoogleContainerRegistryImage(
        name=packageable.executable_spec.name,
        image_path=container.image_path,
        args=packageable.args,
        env_vars=packageable.env_vars,
    )

  instance = docker_adapter.instance()
  client = instance.get_client()
  print(f'Pulling {container.image_path}...')
  repository, tag = instance.split_tag(container.image_path)
  image_id = instance.pull_image(container.image_path)
  image = client.images.get(image_id)

  push_image_tag = _get_push_image_tag(packageable.executor_spec)
  if not push_image_tag:
    if repository.startswith(gcr_project_prefix):
      # If the image path already points to the project's GCR, reuse it.
      push_image_tag = f'{repository}:{tag}'
    else:
      # Otherwise, create a new image repository inside the project's GCR.
      push_image_tag = f'{gcr_project_prefix}/{repository}:{tag}'
  image.tag(push_image_tag)
  print(f'Pushing {push_image_tag}...')
  client.images.push(push_image_tag)
  return local_executables.GoogleContainerRegistryImage(
      name=packageable.executable_spec.name,
      image_path=push_image_tag,
      args=packageable.args,
      env_vars=packageable.env_vars,
  )


def _package_dockerfile(packageable: xm.Packageable, dockerfile: xm.Dockerfile):
  """Matcher method for packaging `xm.Dockerfile`."""
  push_image_tag = _get_push_image_tag(packageable.executor_spec)
  if not push_image_tag:
    gcr_project_prefix = 'gcr.io/' + auth.get_project_name()
    tag = docker_lib.create_tag()
    push_image_tag = f'{gcr_project_prefix}/{dockerfile.name}:{tag}'

  build_image.push(
      build_image.build_by_dockerfile(dockerfile.path, dockerfile.dockerfile,
                                      push_image_tag))
  return local_executables.GoogleContainerRegistryImage(
      name=packageable.executable_spec.name,
      image_path=push_image_tag,
      args=packageable.args,
      env_vars=packageable.env_vars,
  )


def _package_python_container(
    packageable: xm.Packageable,
    python_container: xm.PythonContainer) -> xm.Executable:
  """Matcher method for packaging `xm.PythonContainer`."""
  push_image_tag = _get_push_image_tag(packageable.executor_spec)
  image = build_image.push(
      build_image.build(python_container, packageable.args,
                        packageable.env_vars, push_image_tag))
  return local_executables.GoogleContainerRegistryImage(
      name=packageable.executable_spec.name,
      image_path=image,
  )


def _throw_on_unknown_executable(packageable: xm.Packageable,
                                 executable: xm.ExecutableSpec):
  raise TypeError('Unsupported executable specification '
                  f'for Cloud packaging: {executable!r}')


_CLOUD_PACKAGING_ROUTER = pattern_matching.match(
    _package_container,
    _package_dockerfile,
    _package_python_container,
    _throw_on_unknown_executable,
)


def package_cloud_executable(
    packageable: xm.Packageable,
    executable_spec: xm.ExecutableSpec) -> xm.Executable:
  return _CLOUD_PACKAGING_ROUTER(packageable, executable_spec)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods for routing packageables to appropriate packagers."""

from typing import Any, List, Sequence

from xmanager import xm
from xmanager.xm import pattern_matching
from xmanager.xm_local import executors
from xmanager.xm_local.packaging import cloud as cloud_packaging
from xmanager.xm_local.packaging import local as local_packaging


def _visit_caip_spec(packageable: xm.Packageable, _: executors.CaipSpec):
  return cloud_packaging.package_cloud_executable(packageable,
                                                  packageable.executable_spec)


def _visit_local_spec(packageable: xm.Packageable, _: executors.LocalSpec):
  return local_packaging.package_for_local_executor(packageable,
                                                    packageable.executable_spec)


def _visit_kubernetes_spec(packageable: xm.Packageable,
                           _: executors.KubernetesSpec):
  return cloud_packaging.package_cloud_executable(packageable,
                                                  packageable.executable_spec)


def _throw_on_unknown_executor(packageable: xm.Packageable, executor: Any):
  raise TypeError(f'Unsupported executor specification: {executor!r}. '
                  f'Packageable: {packageable!r}')


_PACKAGING_ROUTER = pattern_matching.match(_visit_caip_spec, _visit_local_spec,
                                           _visit_kubernetes_spec,
                                           _throw_on_unknown_executor)


def package(packageables: Sequence[xm.Packageable]) -> List[xm.Executable]:
  """Routes a packageable to an appropriate packaging mechanism."""
  # TODO: Bundle together.
  return [
      _PACKAGING_ROUTER(packageable, packageable.executor_spec)
      for packageable in packageables
  ]



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for xmanager.xm.resources."""

import unittest
from absl.testing import parameterized

from xmanager import xm
from xmanager.xm import resources
from xmanager.xm.resources import JobRequirements
from xmanager.xm.resources import ResourceType


class ResourceDictTest(unittest.TestCase):

  def test_resource_type_by_name(self):
    self.assertEqual(ResourceType['cpu'], ResourceType.CPU)
    self.assertEqual(ResourceType['Cpu'], ResourceType.CPU)
    self.assertEqual(ResourceType['CPU'], ResourceType.CPU)
    with self.assertRaises(KeyError):
      ResourceType['UPC']  # pylint: disable=pointless-statement

  def test_resource_dict_to_string(self):
    resource_dict = resources.ResourceDict()
    resource_dict[ResourceType.V100] = 8
    resource_dict[ResourceType.CPU] = 4.2 * xm.vCPU
    resource_dict[ResourceType.MEMORY] = 16.5 * xm.GiB

    self.assertEqual(
        str(resource_dict), 'CPU: 4.2, MEMORY: 17716740096.0, V100: 8')

  def test_resource_dict_from_job_requirements(self):
    requirements = JobRequirements(cpu=0.5 * xm.vCPU, memory=2 * xm.MiB, v100=8)
    resource_dict = requirements.task_requirements
    self.assertEqual(resource_dict[ResourceType.CPU], 0.5)
    self.assertEqual(resource_dict[ResourceType.MEMORY], 2097152)
    self.assertEqual(resource_dict[ResourceType.V100], 8)

  def test_job_requirements_unknown_key(self):
    with self.assertRaises(KeyError):
      JobRequirements(cpu=0.5 * xm.vCPU, upc=2)


class TopologyTest(parameterized.TestCase):

  @parameterized.parameters(('2', 2), ('4x4', 16), ('2x3x5', 30),
                            ('4x4_twisted', 16))
  def test_resource_type_by_name(self, topology, chip_count):
    self.assertEqual(resources.Topology(topology).chip_count, chip_count)

  def test_invalid_topology(self):
    with self.assertRaises(resources.InvalidTpuTopologyError):
      resources.Topology('euclidian')

  def test_topology_repr(self):
    self.assertEqual(repr(resources.Topology('4x4')), "xm.Topology('4x4')")

  def test_topology_eq(self):
    self.assertEqual(resources.Topology('4x4'), resources.Topology('4x4'))
    self.assertNotEqual(resources.Topology('2x2'), resources.Topology('4x4'))

    self.assertEqual(
        hash(resources.Topology('4x4')), hash(resources.Topology('4x4')))


class JobRequirementsTest(parameterized.TestCase):

  def test_cpu_job(self):
    requirements = resources.JobRequirements(cpu=1.2, ram=1 * xm.GiB)
    self.assertEqual(requirements.task_requirements[resources.ResourceType.CPU],
                     1.2)
    self.assertEqual(requirements.task_requirements[resources.ResourceType.RAM],
                     1 * xm.GiB)
    self.assertIsNone(requirements.accelerator)
    self.assertIsNone(requirements.topology)
    self.assertEqual(requirements.replicas, 1)

  def test_construct_requirements(self):
    requirements = resources.JobRequirements({resources.ResourceType.CPU: 4},
                                             v100=1)
    task_requirements = requirements.task_requirements
    self.assertEqual(task_requirements[resources.ResourceType.CPU], 4)
    self.assertEqual(task_requirements[resources.ResourceType.V100], 1)

  def test_resource_specified_twice(self):
    with self.assertRaises(ValueError):
      resources.JobRequirements({resources.ResourceType.CPU: 1}, cpu=2)

  def test_tpu_job(self):
    requirements = resources.JobRequirements(tpu_v3='4x4')
    self.assertEqual(requirements.accelerator, resources.ResourceType.TPU_V3)
    self.assertEqual(requirements.topology.name, '4x4')

  def test_location(self):
    requirements = resources.JobRequirements(location='lon_r7')
    self.assertEqual(requirements.location, 'lon_r7')

  def test_service_tier(self):
    requirements = resources.JobRequirements(
        service_tier=resources.ServiceTier.PROD)
    self.assertEqual(requirements.service_tier, resources.ServiceTier.PROD)

  def test_replicas(self):
    requirements = resources.JobRequirements(replicas=2)
    self.assertEqual(requirements.replicas, 2)

    with self.assertRaises(ValueError):
      resources.JobRequirements(replicas=2, tpu_v3='1x1')

  def test_str(self):
    self.assertEqual(
        repr(
            resources.JobRequirements(
                cpu=1,
                location='lon_r7',
                service_tier=resources.ServiceTier.BATCH,
                replicas=2)),
        "xm.JobRequirements(cpu=1.0, location='lon_r7', service_tier=xm.ServiceTier.BATCH, replicas=2)"
    )

  def test_str_omits_empty_fields(self):
    self.assertEqual(
        repr(resources.JobRequirements(cpu=1)), 'xm.JobRequirements(cpu=1.0)')

  def test_is_gpu_tpu_given_cpu(self):
    requirements = resources.JobRequirements(cpu=1, ram=4 * xm.GiB)
    self.assertNotIn(requirements.accelerator, xm.GpuType)
    self.assertNotIn(requirements.accelerator, xm.TpuType)

  def test_is_gpu_tpu_given_gpu(self):
    requirements = resources.JobRequirements(cpu=1, v100=4)
    self.assertIn(requirements.accelerator, xm.GpuType)
    self.assertNotIn(requirements.accelerator, xm.TpuType)

  def test_is_gpu_tpu_given_tpu(self):
    requirements = resources.JobRequirements(cpu=1, tpu_v2='2x2')
    self.assertNotIn(requirements.accelerator, xm.GpuType)
    self.assertIn(requirements.accelerator, xm.TpuType)


class EnumSubsetTest(parameterized.TestCase):

  def test_construction(self):
    self.assertEqual(resources.GpuType['V100'], resources.GpuType(17))
    with self.assertRaises(AttributeError):
      resources.GpuType['TPU_V3']  # pylint: disable=pointless-statement
    with self.assertRaises(ValueError):
      resources.GpuType(170)

  def test_equivalence(self):
    self.assertEqual(resources.ResourceType.V100, resources.GpuType.V100)

  def test_iteration(self):
    gpus = set(iter(resources.GpuType))
    self.assertIn(resources.ResourceType.V100, gpus)
    self.assertNotIn(resources.ResourceType.TPU_V3, gpus)

  def test_contains(self):
    self.assertIn(resources.ResourceType.V100, resources.GpuType)
    self.assertNotIn(resources.ResourceType.TPU_V3, resources.GpuType)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A module for communicating with the Bazel server."""

import abc
import os
import subprocess
from typing import List, Sequence, Tuple

from absl import flags
import attr
from xmanager.bazel import file_utils

from google.protobuf.internal.decoder import _DecodeVarint32
from xmanager.generated import build_event_stream_pb2 as bes_pb2

_BAZEL_COMMAND = flags.DEFINE_string('xm_bazel_command', 'bazel',
                                     'A command that runs Bazel.')


def _get_important_output(events: Sequence[bes_pb2.BuildEvent],
                          label: str) -> Sequence[bes_pb2.File]:
  for event in events:
    if event.id.HasField('target_completed'):
      # Note that we ignore `event.id.target_completed.aspect`.
      if event.id.target_completed.label == label:
        return event.completed.important_output
  raise ValueError(f'Missing target completion event for {label} in Bazel logs')


def _get_normalized_label(events: Sequence[bes_pb2.BuildEvent],
                          label: str) -> str:
  for event in events:
    if event.id.HasField('pattern'):
      assert len(event.id.pattern.pattern) == 1
      if event.id.pattern.pattern[0] == label:
        # Assume there is just one `TargetConfiguredId` child in such events.
        # Note that we ignore `event.children[0].target_configured.aspect`.
        return event.children[0].target_configured.label
  raise ValueError(f'Missing pattern expansion event for {label} in Bazel logs')


def _get_workspace_directory(events: Sequence[bes_pb2.BuildEvent]) -> str:
  for event in events:
    if event.id.HasField('started'):
      return event.started.workspace_directory
  raise ValueError('Missing start event in Bazel logs')


def _read_build_events(path: str) -> List[bes_pb2.BuildEvent]:
  """Parses build events from a file referenced by a given `path`.

  The file should contain serialized length-delimited`bes_pb2.BuildEvent`
  messages. See
  https://docs.bazel.build/versions/master/build-event-protocol.html#consume-in-binary-format
  for details.

  Args:
    path: Path to a file with the protocol.

  Returns:
    A list of build events.
  """
  with open(path, 'rb') as bep_file:
    buffer = bep_file.read()
    events = []
    position = 0
    while position < len(buffer):
      # Reimplementation of Java's `AbstractParser.parseDelimitedFrom` for
      # protobufs, which is not available in Python.
      size, start = _DecodeVarint32(buffer, position)
      event = bes_pb2.BuildEvent()
      event.ParseFromString(buffer[start:start + size])
      events.append(event)
      position = start + size
    return events


def _root_absolute_path() -> str:
  # If the launch script is run with Bazel, use `BUILD_WORKSPACE_DIRECTORY` to
  # get the root of the workspace where the build was initiated. If the launch
  # script is run with the CLI, query Bazel to find out.
  return os.getenv('BUILD_WORKSPACE_DIRECTORY') or subprocess.run(
      [_BAZEL_COMMAND.value, 'info', 'workspace'],
      check=True,
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
      universal_newlines=True,
  ).stdout.strip()


def _should_wrap_in_par(label: str) -> bool:
  """Returns whether a build target needs to be wrapped in PAR."""
  if label.endswith('.par'):
    return False

  # For each matching target blaze query would produce following line:
  # <rule name> rule <target name>
  # For example:
  # py_library rule //third_party/py/xmanager/xm:__init__
  output = subprocess.run(
      [_BAZEL_COMMAND.value, 'query', label, '--output', 'label_kind'],
      check=True,
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
      cwd=_root_absolute_path(),
  ).stdout.decode('utf-8')
  rule_line = next(filter(lambda line: label in line, output.split(os.linesep)))
  rule = rule_line.split()[0]

  return rule in ('py_binary',)


def build_single_target(label: str, tail_args: Sequence[str] = ()) -> List[str]:
  """Builds a target and returns paths to its important output.

  The definition of 'important artifacts in an output group' can be found at
  https://github.com/bazelbuild/bazel/blob/8346ea4cfdd9fbd170d51a528fee26f912dad2d5/src/main/java/com/google/devtools/build/lib/analysis/TopLevelArtifactHelper.java#L223-L224.

  Args:
    label: Label of the target to build.
    tail_args: Arguments to append to the Bazel command.

  Returns:
    A list of paths to the output.
  """
  if _should_wrap_in_par(label):
    label += '.par'

  with file_utils.TemporaryFilePath() as bep_path:
    subprocess.run(
        [
            _BAZEL_COMMAND.value,
            'build',
            f'--build_event_binary_file={bep_path}',
            # Forces a GC at the end of the build and publishes value to BEP.
            '--memory_profile=/dev/null',
            label,
        ] + list(tail_args),
        check=True,
        cwd=_root_absolute_path(),
    )
    events = _read_build_events(bep_path)
    normalized_label = _get_normalized_label(events, label)
    files = _get_important_output(events, normalized_label)
    workspace = _get_workspace_directory(events)
    return [
        os.path.join(workspace, *file.path_prefix, file.name) for file in files
    ]


class BazelService(abc.ABC):
  """An interface for Bazel operations."""

  @abc.abstractmethod
  def fetch_kinds(self, labels: Sequence[str]) -> List[str]:
    """Fetches kinds of given targets.

    See https://docs.bazel.build/versions/main/query.html#output-label_kind.

    Args:
      labels: Labels of the targets to query.

    Returns:
      A list of kinds, for example, `['py_binary rule']`.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def build_targets(self, labels: Sequence[str],
                    tail_args: Sequence[str]) -> List[List[str]]:
    """Builds given targets and returns paths to their important outputs.

    Args:
      labels: Labels of the targets to build.
      tail_args: Arguments to append to the Bazel command.

    Returns:
      For each label returns a list of its important outputs.
    """
    raise NotImplementedError


def _to_tuple(sequence: Sequence[str]) -> Tuple[str, ...]:
  """A standalone function to satisfy PyType."""
  return tuple(sequence)


@attr.s(auto_attribs=True, frozen=True)
class BazelTarget:
  """A Bazel target to be built."""

  label: str
  bazel_args: Tuple[str, ...] = attr.ib(converter=_to_tuple)



"""An utility to generate factory methods in packagables.py.

packagables.py has a set of convenience functions which can be mechanically
generated for each ExecutableSpec. But we want them to have all arguments
explicetly written down with proper type annotations. This way IDEs can provide
proper contextual help and autocompletion. We also want make it easy to find
documentation, so having docstrings inplace is important.

We do this in an automated manner to ensure that arguments in ExecutableSpecs
and in the packagables do not diverge and that documentation remains up to date.

Usage: Run this binary and replace relevant parts of packagables.py. Then run
pyformat.
"""

import inspect
import re
from typing import List, Sequence, Type

from absl import app
import inflection
from xmanager.xm import executables
from xmanager.xm import job_blocks

_EXECUTABLES_SPECS = (
    executables.Binary,
    executables.BazelBinary,
    executables.Container,
    executables.BazelContainer,
    executables.PythonContainer,
    executables.Dockerfile,
)

# Reconstructing argument definition from Python introspection is non trivial.
# Partially due to deducing shorter module names (job_blocks rather than
# xmanager.xm.job_blocks). But mostly because type annotations for attr.s
# constructor show field types rather than what converters accept.
# So here we just list how each parameter should be defined.
_KNOWN_ARGS = (
    'base_image: Optional[str] = None',
    'bazel_args: Collection[str] = ()',
    'docker_instructions: Optional[List[str]] = None',
    'dockerfile: Optional[str] = None',
    'entrypoint: Union[executables.ModuleName, executables.CommandList]',
    'dependencies: Collection[executables.BinaryDependency] = ()',
    'image_path: str',
    'label: str',
    'path: str',
    'use_deep_module: bool',
)
_KNOWN_ARGS_DICT = {arg.split(':')[0]: arg for arg in _KNOWN_ARGS}

_ATTRIBUTES_SECTION_HEADER = '  Attributes:'
_ARGS_DOCSTRING = """  Args:
    executor_spec: Where the binary should be launched. Instructs for which
      platform it should be packaged."""
_DOCSTRING_SUFFIX = """
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as two arguments: --batch_size 16. If
      positional arguments are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to be set.

  Returns:
    A packageable object which can be turned into an executable with
    Experiment.package or Experiment.package_async.
"""


def generate_docstring(executable: Type[job_blocks.ExecutableSpec]) -> str:
  """Returns a docstring for a ExecutableSpec factory method."""
  docstring = executable.__doc__
  if _ATTRIBUTES_SECTION_HEADER not in docstring:
    raise Exception(
        f'Please add Attributes: section to {executable.__name__} docstring.')
  docstring = re.sub(_ATTRIBUTES_SECTION_HEADER, _ARGS_DOCSTRING, docstring)
  docstring = docstring.rstrip() + _DOCSTRING_SUFFIX.rstrip()
  return docstring


def generate_factory_parameters(parameters: List[inspect.Parameter]) -> str:
  """Returns ExecutableSpec factory method parameters definition.

  Args:
    parameters: ExecutableSpec constructor parameters except self.

  Returns:
    Python source code.
  """
  source = '    executor_spec: job_blocks.ExecutorSpec,\n'

  keyword_args_started = False
  for parameter in parameters:
    if (parameter.kind == inspect.Parameter.KEYWORD_ONLY and
        not keyword_args_started):
      keyword_args_started = True
      source += '    *,\n'

    parameter_source = _KNOWN_ARGS_DICT[parameter.name]
    if (parameter.default != inspect.Parameter.empty and
        '=' not in parameter_source):
      parameter_source += f' = {parameter.default!r}'

    source += f'    {parameter_source},\n'

  if not keyword_args_started:
    source += '    *,\n'

  source += '    args: Optional[job_blocks.UserArgs] = None,\n'
  source += '    env_vars: Mapping[str, str] = immutabledict.immutabledict(),\n'
  return source


def generate_factory_method(executable: Type[job_blocks.ExecutableSpec]) -> str:
  """Returns the factory method source code for the given ExecutableSpec."""

  factory_name = inflection.underscore(executable.__name__)
  if factory_name == 'dockerfile':
    # We should rename the Dockerfile class to avoid this naming inconsistency.
    factory_name = 'dockerfile_container'

  signature = inspect.signature(executable.__init__)
  # Skip the `self` parameter.
  parameters = list(signature.parameters.values())[1:]

  executable_args = '\n'.join(
      f'          {p.name}={p.name},' for p in parameters)

  return f'''
def {factory_name}(
{generate_factory_parameters(parameters)}
) -> job_blocks.Packageable:
  # pyformat: disable
  """{generate_docstring(executable)}
  """
  # pyformat: enable
  return job_blocks.Packageable(
      executable_spec=executables.{executable.__name__}(
{executable_args}
      ),
      executor_spec=executor_spec,
      args=args,
      env_vars=env_vars,
  )'''.strip('\n')


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  for spec in _EXECUTABLES_SPECS:
    print(generate_factory_method(spec))
    print()
    print()


if __name__ == '__main__':
  app.run(main)



from absl.testing import absltest
from xmanager import xm
from xmanager.cloud import build_image


class BuildImageTest(absltest.TestCase):

  def create_container(self, entrypoint) -> xm.PythonContainer:
    return xm.PythonContainer(entrypoint=entrypoint)

  def test_get_entrypoint_commands_module_adds_suffix(self):
    project = self.create_container(xm.ModuleName('some.python.module'))
    entrypoint_commands = build_image._get_entrypoint_commands(project)
    self.assertEndsWith(entrypoint_commands, ' "$@"')

  def test_get_entrypoint_commands_adds_suffix(self):
    commands = ['echo "aaa"']
    project = self.create_container(xm.CommandList(commands))
    entrypoint_commands = build_image._get_entrypoint_commands(project)
    self.assertEndsWith(entrypoint_commands, ' "$@"')

  def test_get_entrypoint_commands_no_dup_plain_suffix(self):
    commands = ['echo "aaa" $@']
    project = self.create_container(xm.CommandList(commands))
    entrypoint_commands = build_image._get_entrypoint_commands(project)
    self.assertEndsWith(entrypoint_commands, ' $@')

  def test_get_entrypoint_commands_no_dup_quoted_suffix(self):
    commands = ['echo "aaa" "$@"']
    project = self.create_container(xm.CommandList(commands))
    entrypoint_commands = build_image._get_entrypoint_commands(project)
    self.assertEndsWith(entrypoint_commands, ' "$@"')
    self.assertNotEndsWith(entrypoint_commands, ' "$@" "$@"')

  def test_get_entrypoint_commands_dup_single_quoted_suffix(self):
    commands = ['echo "aaa" \'$@\'']
    project = self.create_container(xm.CommandList(commands))
    entrypoint_commands = build_image._get_entrypoint_commands(project)
    self.assertEndsWith(entrypoint_commands, ' \'$@\' "$@"')


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods for executing job groups using the local backend."""

import abc
import asyncio
import atexit
from concurrent import futures
import threading
from typing import Any, Awaitable, Callable, List, cast

from absl import logging
import attr
from docker.models import containers
from xmanager import xm
from xmanager.docker import docker_adapter
from xmanager.xm import job_operators
from xmanager.xm import pattern_matching
from xmanager.xm import utils
from xmanager.xm_local import executables
from xmanager.xm_local import executors
from xmanager.xm_local import status

_DEFAULT_ENCODING = 'utf-8'
_BRIDGE_NETWORK_NAME = 'xmanager'


def _print_chunk(name: str, line: str) -> None:
  print('[{}] {}'.format(name, line.strip()))


class ExecutionHandle(abc.ABC):
  """An interface for operating on executions."""

  @abc.abstractmethod
  async def wait(self) -> None:
    raise NotImplementedError

  @abc.abstractmethod
  def get_status(self) -> status.LocalWorkUnitStatus:
    """Aggregates the statuses of all jobs in the work unit into one status."""
    raise NotImplementedError


class LocalExecutionHandle(ExecutionHandle, abc.ABC):
  """An interface for operating on local executions."""

  @abc.abstractmethod
  async def monitor(self) -> None:
    raise NotImplementedError

  @abc.abstractmethod
  def terminate(self) -> None:
    raise NotImplementedError


async def _throw_on_unknown_executable(
    get_full_job_name: Callable[[str], str],
    job: xm.Job,
    executable: Any,
) -> LocalExecutionHandle:
  raise TypeError(f'Unsupported executable for local execution: {executable!r}')


@attr.s(auto_attribs=True)
class ContainerHandle(LocalExecutionHandle):
  """A handle for referring to the launched container."""

  name: str
  model: containers.Container
  stream_output: bool
  futures_executor: futures.Executor = attr.Factory(futures.ThreadPoolExecutor)

  async def wait(self) -> None:
    response = await asyncio.wrap_future(
        self.futures_executor.submit(self.model.wait))
    status_code = response['StatusCode']
    if status_code != 0:
      raise RuntimeError(
          f'Container {self.model!r} returned non-zero status: {status_code}')

  def get_status(self) -> status.LocalWorkUnitStatus:
    raise NotImplementedError

  def terminate(self) -> None:
    self.model.stop()
    self.futures_executor.shutdown(wait=True)

  async def monitor(self) -> None:

    def _stream_chunks() -> None:
      for chunk in self.model.logs(stream=True, follow=True):
        _print_chunk(self.name, chunk.decode(_DEFAULT_ENCODING))

    if self.stream_output:
      await asyncio.wrap_future(self.futures_executor.submit(_stream_chunks))


async def _launch_loaded_container_image(
    get_full_job_name: Callable[[str], str],
    job: xm.Job,
    executable: executables.LoadedContainerImage,
) -> LocalExecutionHandle:
  """Launches a preloaded image as a detached container."""
  if not isinstance(job.executor, executors.Local):
    raise TypeError(f'Expected {job!r} to have the Local executor')
  executor = cast(executors.Local, job.executor)
  instance = docker_adapter.instance()

  if not instance.has_network(_BRIDGE_NETWORK_NAME):
    instance.create_network(_BRIDGE_NETWORK_NAME)

  args = xm.merge_args(executable.args, job.args).to_list(utils.ARG_ESCAPER)
  env_vars = {**executable.env_vars, **job.env_vars}
  options = executor.docker_options or executors.DockerOptions()

  container = instance.run_container(
      name=get_full_job_name(job.name),
      image_id=executable.image_id,
      network=_BRIDGE_NETWORK_NAME,
      args=args,
      env_vars=env_vars,
      ports=options.ports or {},
      volumes=options.volumes or {},
  )
  return ContainerHandle(
      name=job.name,
      model=container,
      stream_output=executor.experimental_stream_output,
  )


@attr.s(auto_attribs=True)
class BinaryHandle(LocalExecutionHandle):
  """A handle referring to the launched binary."""

  name: str
  process: asyncio.subprocess.Process
  stream_output: bool

  async def wait(self) -> None:
    return_code = await self.process.wait()
    if return_code != 0:
      raise RuntimeError(
          f'Process {self.process!r} returned non-zero code: {return_code}')

  def get_status(self) -> status.LocalWorkUnitStatus:
    raise NotImplementedError

  def terminate(self) -> None:
    self.process.terminate()

  async def monitor(self) -> None:
    if self.stream_output:
      while True:
        line = await self.process.stdout.readline()
        if not line:
          break
        _print_chunk(self.name, line.decode(_DEFAULT_ENCODING))


async def _launch_local_binary(
    get_full_job_name: Callable[[str], str],
    job: xm.Job,
    executable: executables.LocalBinary,
) -> LocalExecutionHandle:
  """Launches a local binary as a detached process."""
  del get_full_job_name  # Unused.
  if not isinstance(job.executor, executors.Local):
    raise TypeError(f'Expected {job!r} to have the Local executor')

  args = xm.merge_args(executable.args, job.args).to_list(utils.ARG_ESCAPER)
  env_vars = {**executable.env_vars, **job.env_vars}
  process = await asyncio.create_subprocess_exec(
      executable.path,
      *args,
      env=env_vars,
      start_new_session=True,
      stdout=asyncio.subprocess.PIPE
      if job.executor.experimental_stream_output else None,
      stderr=asyncio.subprocess.STDOUT
      if job.executor.experimental_stream_output else None,
  )
  return BinaryHandle(
      name=job.name,
      process=process,
      stream_output=job.executor.experimental_stream_output,
  )


# PyType infers the return type of `async` functions without wrapping them with
# `Awaitable`, so we are overriding the type of `_LOCAL_EXECUTION_ROUTER` to
# make it right.
_LocalExecutionRouter = Callable[[Callable[[str], str], xm.Job, Any],
                                 Awaitable[LocalExecutionHandle]]
_LOCAL_EXECUTION_ROUTER: _LocalExecutionRouter = pattern_matching.match(
    _launch_loaded_container_image,
    _launch_local_binary,
    _throw_on_unknown_executable,  # pytype: disable=annotation-type-mismatch
)

# Note that currently handles are never removed from the list. We can consider
# removing them on completion if needed.
_local_jobs: List[LocalExecutionHandle] = []
_local_jobs_lock = threading.Lock()


@atexit.register
def _terminate_local_jobs():
  """Terminates local jobs that were launched during the current session."""
  with _local_jobs_lock:
    if _local_jobs:
      print(f'Terminating {len(_local_jobs)} local job(s)'
            ' that may still be running...')
    for local_job in _local_jobs:
      try:
        local_job.terminate()
      except Exception:  # pylint: disable=broad-except
        logging.warning('Unable to terminate %s', repr(local_job))


def _local_job_predicate(job: xm.Job) -> bool:
  return isinstance(job.executor, executors.Local)


async def launch(get_full_job_name: Callable[[str], str],
                 job_group: xm.JobGroup) -> List[LocalExecutionHandle]:
  """Launches jobs with `xm_local.Local` executor."""
  # Must act on all jobs with `Local` executor.
  local_jobs = job_operators.collect_jobs_by_filter(job_group,
                                                    _local_job_predicate)
  handles: List[LocalExecutionHandle] = [
      await _LOCAL_EXECUTION_ROUTER(get_full_job_name, job, job.executable)
      for job in local_jobs
  ]
  with _local_jobs_lock:
    _local_jobs.extend(handles)
  return handles



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper methods to find the cluster details when running on CAIP."""

import json
import os
import re
from typing import List, Sequence, Tuple
import urllib.request

from absl import logging
from google.cloud import storage


def get_master_address_port() -> Tuple[str, str]:
  """Get the master worker from CLUSTER_SPEC.

  https://cloud.google.com/ai-platform/training/docs/distributed-training-containers#about-cluster-spec

  Returns:
    address string and port string
  """
  cluster_spec = os.environ.get('CLUSTER_SPEC', None)
  logging.info('CLUSTER_SPEC: %s', cluster_spec)
  if not cluster_spec:
    return '127.0.0.1', '29500'

  cluster_spec = json.loads(cluster_spec)
  master = cluster_spec['cluster']['workerpool0'][0]
  [addr, port] = master.split(':')
  return addr, port


def get_world_size_rank() -> Tuple[int, int]:
  """Get the world size and rank of the current replica from CLUSTER_SPEC."""
  cluster_spec = os.environ.get('CLUSTER_SPEC', None)
  if not cluster_spec:
    return 1, 0

  cluster_spec = json.loads(cluster_spec)
  world_size = 0
  for pool in cluster_spec['cluster']:
    if pool == cluster_spec['task']['type']:
      rank = world_size + cluster_spec['task']['index']
    world_size += len(cluster_spec['cluster'][pool])

  logging.info('WORLD SIZE: %d; RANK: %d', world_size, rank)
  return world_size, rank


def create_cluster_specs(workers: Sequence[str]) -> List[str]:
  """Takes a list of domain names and constructs a CLUSTER_SPEC for each."""
  cluster = {}
  for i, domain in enumerate(workers):
    cluster[f'workerpool{i}'] = [domain]
  specs = []
  for i in range(len(workers)):
    spec = {
        'cluster': cluster,
        'task': {
            'type': f'workerpool{i}',
            'index': i
        },
    }
    specs.append(json.dumps(spec))
  return specs


def get_workerpool_address(workerpool: str) -> str:
  """Creates a late-binding that is mapped at runtime."""
  return f'%objectname({workerpool})%'


def map_workerpool_address_args(args: List[str]) -> List[str]:
  """Maps late-binding to workerpool addresses at runtime."""
  cluster_spec = os.environ.get('CLUSTER_SPEC')
  if cluster_spec is None:
    return args

  # capture %objectname(<capture-group>)%
  late_bind_regex = re.compile(r'\%objectname\((.*)\)\%')
  cluster_spec = json.loads(cluster_spec)['cluster']
  result = []
  for arg in args:
    match = late_bind_regex.search(arg)
    if match is None:
      result.append(arg)
    else:
      worker_type = match.group(1)
      result.append(
          arg.replace(f'%objectname({worker_type})%',
                      cluster_spec[worker_type][0]))
  return result


def print_workerpool_address_args(argv: List[str]) -> None:
  # Note that this is method is called by
  # third_party/py/xmanager/cloud/data/wrapped_entrypoint.sh
  for arg in map_workerpool_address_args(argv[1:]):
    print(arg,)


def create_workerpool_address_env_vars_script(path: str) -> None:
  """Create a script to map late-binding env vars to their value at runtime."""
  with open(path, 'w') as f:
    f.write('#!/bin/bash\n\n')

  cluster_spec = os.environ.get('CLUSTER_SPEC', None)
  if cluster_spec is None:
    return

  content = []
  # capture %objectname(<capture-group>)%
  late_bind_regex = re.compile(r'\%objectname\((.*)\)\%')
  cluster_spec = json.loads(cluster_spec)['cluster']
  for key, value in os.environ.items():
    match = late_bind_regex.match(value)
    if match:
      worker_type = match.group(1)
      content.append(f'export {key}={cluster_spec[worker_type][0]}')

  if content:
    with open(path, 'a') as f:
      f.write('\n'.join(content))


def get_region() -> str:
  """Get the region of the current GCE VM from metadata, e.g. us-central1."""
  # Default VM instance metadata
  # https://cloud.google.com/compute/docs/metadata/default-metadata-values#vm_instance_metadata
  request = urllib.request.Request(
      'http://metadata.google.internal/computeMetadata/v1/instance/zone')
  request.add_header('Metadata-Flavor', 'Google')
  response = urllib.request.urlopen(request)
  content = str(response.read())
  zone = content.strip('\'').split('/')[-1]
  region = zone[:-2]
  return region


def get_closest_bucket(bucket_names: List[str]) -> str:
  """Get the closest bucket from a list of buckets."""
  region = get_region()
  for name in bucket_names:
    b = storage.Bucket(client=storage.Client(), name=name)
    b.reload()
    # Only works for REGIONAL and MULT_REGIONAL.
    if region.startswith(b.location.lower()):
      return name
  logging.info('None of the buckets are close. Picking a random one.')
  return bucket_names[0]



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convenience adapter for the standard client."""

import functools
from typing import Dict, List, Mapping, Sequence, Tuple, Union

from absl import logging
import docker
from docker import errors
from docker.models import containers
from docker.utils import utils

Ports = Dict[Union[int, str], Union[None, int, Tuple[str, int], List[int]]]


@functools.lru_cache()
def instance() -> 'DockerAdapter':
  """Returns a thread-safe singleton adapter derived from the environment.

  Allows the user to ignore the complexities of the underlying library, and
  focus on a concrete small subset of required actions.
  """
  return DockerAdapter(docker.from_env())


class DockerAdapter(object):
  """Convenience adapter for the standard client."""

  def __init__(self, client: docker.DockerClient) -> None:
    self._client = client

  def has_network(self, name: str) -> bool:
    return bool(self._client.networks.list([name]))

  def create_network(self, name: str) -> str:
    return self._client.networks.create(name).id

  def get_client(self) -> docker.DockerClient:
    return self._client

  def is_registry_label(self, label: str) -> bool:
    try:
      self._client.images.get_registry_data(label)
      return True
    except errors.NotFound:
      return False

  def split_tag(self, image_tag: str) -> Tuple[str, str]:
    repository, tag = utils.parse_repository_tag(image_tag)
    return repository, tag or 'latest'

  def pull_image(self, image_tag: str) -> str:
    repository, tag = self.split_tag(image_tag)
    # Without a tag, Docker will try to pull every image instead of latest.
    # From docker>=4.4.0, use `client.image.pull(*args, all_tags=False)`.
    return self._client.images.pull(repository, tag=tag).id

  def load_image(self, path: str) -> str:
    with open(path, 'rb') as data:
      images = self._client.images.load(data)
      if len(images) != 1:
        raise ValueError(f'{path} must contain precisely one image')
      return images[0].id

  def run_container(
      self,
      name: str,
      image_id: str,
      args: Sequence[str],
      env_vars: Mapping[str, str],
      network: str,
      ports: Ports,
      volumes: Dict[str, str],
  ) -> containers.Container:
    """Runs a given container image."""
    make_mount = lambda guest: {'bind': guest, 'mode': 'rw'}
    return self._client.containers.run(
        image_id,
        name=name,
        hostname=name,
        network=network,
        detach=True,
        remove=True,
        command=args,
        environment=env_vars,
        ports=ports,
        volumes={host: make_mount(guest) for host, guest in volumes.items()},
    )

  def stop_container(self, container_id: str) -> None:
    try:
      self._client.containers.get(container_id).stop()
    except docker.errors.NotFound:
      logging.warning(
          'Container %s could not be stopped as it was not found '
          '(it may already have been stopped)', container_id)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common job operators useful in the framework and 3P-libraries."""

import copy
import itertools
from typing import Any, Callable, List, Sequence, Tuple

import attr
from xmanager.xm import job_blocks
from xmanager.xm import pattern_matching


def shallow_copy_job_type(
    job_type: job_blocks.JobTypeVar) -> job_blocks.JobTypeVar:
  """Creates a shallow copy of the job structure."""

  def apply_to_job_group(job_group: job_blocks.JobGroup) -> job_blocks.JobGroup:
    job_group = copy.copy(job_group)
    job_group.jobs = {key: matcher(job) for key, job in job_group.jobs.items()}
    return job_group

  matcher = pattern_matching.match(
      pattern_matching.Case([job_blocks.Job], copy.copy),
      apply_to_job_group,
      pattern_matching.Case([job_blocks.JobGeneratorType],
                            lambda generator: generator),
  )
  return matcher(job_type)


def populate_job_names(job_type: job_blocks.JobTypeVar) -> None:
  """Assigns default names to the given jobs."""

  def apply_to_job(prefix: Sequence[str], target: job_blocks.Job) -> None:
    if target.name is None:
      target.name = '_'.join(prefix) if prefix else target.executable.name

  def apply_to_job_group(prefix: Sequence[str],
                         target: job_blocks.JobGroup) -> None:
    for key, job in target.jobs.items():
      matcher([*prefix, key], job)

  def ignore_unknown(_: Sequence[str], target: Any) -> None:
    return target

  matcher = pattern_matching.match(
      apply_to_job,
      apply_to_job_group,
      ignore_unknown,
  )
  return matcher([], job_type)


def collect_jobs_by_filter(
    job_group: job_blocks.JobGroup,
    predicate: Callable[[job_blocks.Job], bool],
) -> List[job_blocks.Job]:
  """Flattens a given job group and filters the result."""

  def match_job(job: job_blocks.Job) -> List[job_blocks.Job]:
    return [job] if predicate(job) else []

  def match_job_group(job_group: job_blocks.JobGroup) -> List[job_blocks.Job]:
    return list(
        itertools.chain.from_iterable(
            [job_collector(job) for job in job_group.jobs.values()]))

  job_collector = pattern_matching.match(match_job_group, match_job)
  return job_collector(job_group)


@attr.s(auto_attribs=True)
class ConstraintClique:
  """A constraint with the list of jobs it applies to."""

  constraint: job_blocks.Constraint
  jobs: List[job_blocks.Job]


def aggregate_constraint_cliques(
    job_group: job_blocks.JobGroup) -> List[ConstraintClique]:
  """Forms constraint cliques.

  For each constraint met, collects all jobs it applies to.

  Args:
    job_group: A job group to aggregate on.

  Returns:
    A set of cliques.
  """

  def match_job(
      job: job_blocks.Job
  ) -> Tuple[List[ConstraintClique], List[job_blocks.Job]]:
    return [], [job]

  def match_job_group(
      job_group: job_blocks.JobGroup
  ) -> Tuple[List[ConstraintClique], List[job_blocks.Job]]:
    cliques: List[ConstraintClique] = []
    jobs: List[job_blocks.Job] = []
    for job in job_group.jobs.values():
      subcliques, subjobs = matcher(job)
      cliques += subcliques
      jobs += subjobs
    cliques = [
        ConstraintClique(constraint, jobs)
        for constraint in job_group.constraints
    ] + cliques
    return cliques, jobs

  matcher = pattern_matching.match(match_job, match_job_group)
  result, _ = matcher(job_group)
  return result


def flatten_jobs(job_group: job_blocks.JobGroup) -> List[job_blocks.Job]:
  return collect_jobs_by_filter(job_group, lambda _: True)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers to support GCS-based output dir.

For the code running on Google Cloud a GCS directory should be used to store
data.

A GCS name should have a particular format 'gs://bucket/directory'. For XCloud
we provide a shared bucket which can be used by all researchers. To ensure the
format of the name is correct, and to suggest reasonable defaults for it, this
file provides a flag and several helper functions.
"""

import datetime
import getpass
import os

from absl import app
from absl import flags

_GCS_PATH = flags.DEFINE_string(
    'xm_gcs_path', None, 'A GCS directory within a bucket to store output '
    '(in gs://bucket/directory format).')

_GS_PREFIX = 'gs://'
_GCS_PREFIX = '/gcs/'


def suggestion(project_name: str) -> str:
  """Returns a suggested GCS dir name for the given @project_name."""
  return os.path.join(
      _GS_PREFIX, 'xcloud_public_bucket', getpass.getuser(),
      project_name + '-' + datetime.datetime.now().strftime('%Y%m%d-%H%M%S'))


def get_gcs_path_or_suggestion(project_name: str) -> str:
  """Returns a value passed in the --xm_gcs_path flag or a suggested valid path.

  Args:
    project_name: a project name used to generate suggested GCS path.

  Returns:
    If the --xm_gcs_path flag is empty, returns a reasonable suggestion.
    If the --xm_gcs_path contains invalid value, raise an error.
    Otherwise, returns a flag value.
  """
  if not _GCS_PATH.value:
    return suggestion(project_name)
  elif not is_gcs_path(_GCS_PATH.value):
    raise app.UsageError(
        '--xm_gcs_path not in gs://bucket/directory or /gcs/path format. ' +
        f'Suggestion: --xm_gcs_path={suggestion(project_name)}')
  return str(_GCS_PATH.value)


def get_gcs_path_or_fail(project_name: str) -> str:
  """Returns value passed in the --xm_gcs_path flag; fails if nothing is passed.

  Args:
    project_name: a project name used to generate suggested GCS path.

  Returns:
    If the --xm_gcs_path flag is empty, or contains invalid value, raise an
    error. Otherwise, returns a flag value.
  """
  if not _GCS_PATH.value:
    raise app.UsageError('--xm_gcs_path is missing. Suggestion: ' +
                         f'--xm_gcs_path={suggestion(project_name)}')
  elif not is_gcs_path(_GCS_PATH.value):
    raise app.UsageError(
        '--xm_gcs_path not in gs://bucket/directory or /gcs/path format. ' +
        f'Suggestion: --xm_gcs_path={suggestion(project_name)}')
  return str(_GCS_PATH.value)


def is_gs_path(path: str) -> bool:
  """Given the path, checks whether it is a valid Google Storage URL.

  Args:
    path: a path.

  Returns:
    True iff a path starts with the 'gs://' prefix.
  """
  return path.startswith(_GS_PREFIX)


def is_gcs_fuse_path(path: str) -> bool:
  """Given the path, checks whether it is a valid gcs_fuse path.

  Args:
    path: a path.

  Returns:
    True iff a path starts with the '/gcs/' prefix.
  """
  return path.startswith(_GCS_PREFIX)


def is_gcs_path(path: str) -> bool:
  """Given the path, checks whether it is a valid GCS URL supported by XCloud.

  Args:
    path: a path.

  Returns:
    True iff a path starts with either 'gs://' or '/gcs/' prefix.
  """
  return is_gs_path(path) or is_gcs_fuse_path(path)


def _gcs_path_no_prefix(path: str) -> str:
  """Given the GCS path in gs://bucket/directory format, strips 'gs://' from it.

  Args:
    path: GCS path in gs://bucket/directory format.

  Returns:
    Path without 'gs://' prefix.
  """
  if is_gs_path(path):
    return path[len(_GS_PREFIX):]
  if is_gcs_fuse_path(path):
    return path[len(_GCS_PREFIX):]
  raise ValueError(
      f'Path not in gs://bucket/directory or /gcs/path format: {path}')


# Exposed for testing.
gcp_website_url = 'https://console.cloud.google.com'


def get_gcs_url(path: str) -> str:
  """Given the GCS path, provides a GCS URL to access it.

  Args:
    path: GCS path in gs://bucket/directory format.

  Returns:
    GCS URL to access path.
  """
  no_prefix = _gcs_path_no_prefix(path)
  return f'{gcp_website_url}/storage/browser/{no_prefix}'


def get_gcs_fuse_path(path: str) -> str:
  """Given the GCS path, provides corresponding Cloud Storage FUSE path.

  See https://cloud.google.com/storage/docs/gcs-fuse for details.

  Args:
    path: GCS path in gs://bucket/directory format.

  Returns:
    Path in the /gcs/bucket/directory format.
  """
  return _GCS_PREFIX + _gcs_path_no_prefix(path)


def get_gs_path(path: str) -> str:
  """Given the GCS path, provides corresponding Cloud Storage URL.

  See https://cloud.google.com/storage/docs/gcs-fuse for details.

  Args:
    path: GCS path in /gcs/bucket/directory format.

  Returns:
    Path in the gs://bucket/directory format.
  """
  return _GS_PREFIX + _gcs_path_no_prefix(path)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for executables."""

import os
import unittest

from xmanager.xm import executables
from xmanager.xm import utils


class ExecutablesTest(unittest.TestCase):

  def test_python_container_name(self):
    executable = executables.PythonContainer(
        entrypoint=executables.ModuleName('module'),
        path='/home/user/project/',
    )

    self.assertEqual(executable.name, 'project')

  def test_container_name(self):
    executable = executables.Container(
        image_path='/home/user/project/image.tar')

    self.assertEqual(executable.name, 'image_tar')

  def test_binary_name(self):
    executable = executables.Binary(path='./binary')

    self.assertEqual(executable.name, 'binary')

  def test_bazel_container_name(self):
    executable = executables.BazelContainer(label='//container')

    self.assertEqual(executable.name, 'container')

  def test_bazel_binary_name(self):
    executable = executables.BazelBinary(label=':binary')

    self.assertEqual(executable.name, '_binary')

  def test_dockerfile_defaults(self):
    root = utils.resolve_path_relative_to_launcher('.')

    spec = executables.Dockerfile()
    self.assertEqual(spec.path, root)
    self.assertEqual(spec.dockerfile, os.path.join(root, 'Dockerfile'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for testing core objects."""

from concurrent import futures
from typing import Any, Awaitable, Callable, List, Mapping

import attr
from xmanager.xm import core
from xmanager.xm import id_predictor
from xmanager.xm import job_blocks


class TestWorkUnit(core.WorkUnit):
  """A test version of WorkUnit with abstract methods implemented."""

  def __init__(
      self,
      experiment: core.Experiment,
      work_unit_id_predictor: id_predictor.Predictor,
      create_task: Callable[[Awaitable[Any]], futures.Future],
      launched_jobs: List[job_blocks.JobType],
      launched_jobs_args: List[Mapping[str, Any]],
      args: Mapping[str, Any],
  ) -> None:
    super().__init__(experiment, create_task, args, core.WorkUnitRole())
    self._launched_jobs = launched_jobs
    self._launched_jobs_args = launched_jobs_args
    self._work_unit_id = work_unit_id_predictor.reserve_id()

  async def _wait_until_complete(self) -> None:
    """Test work unit is immediately complete."""

  async def _launch_job_group(self, job_group: job_blocks.JobGroup,
                              args: Mapping[str, Any]) -> None:
    """Appends the job group to the launched_jobs list."""
    self._launched_jobs.extend(job_group.jobs.values())
    self._launched_jobs_args.append(args)

  @property
  def work_unit_id(self) -> int:
    return self._work_unit_id

  @property
  def experiment_unit_name(self) -> str:
    return f'{self.experiment_id}_{self._work_unit_id}'


class TestExperiment(core.Experiment):
  """A test version of Experiment with abstract methods implemented."""

  constraints: List[job_blocks.JobType]

  def __init__(self) -> None:
    super().__init__()
    self.launched_jobs = []
    self.launched_jobs_args = []
    self._work_units = []

  def _create_experiment_unit(self, args,
                              role=core.WorkUnitRole()) -> TestWorkUnit:
    """Creates a new WorkUnit instance for the experiment."""
    work_unit = TestWorkUnit(self, self._work_unit_id_predictor,
                             self._create_task, self.launched_jobs,
                             self.launched_jobs_args, args)
    self._work_units.append(work_unit)
    return work_unit

  @property
  def work_unit_count(self) -> int:
    return len(self.work_units)

  @property
  def work_units(self):
    return self._work_units

  @property
  def experiment_id(self) -> int:
    return 1


class TestExecutable(job_blocks.Executable):
  """A test version of Executable with abstract methods implemented."""
  counter = 0

  def __init__(self):
    super().__init__(name=f'{TestExecutable.counter}')
    TestExecutable.counter += 1


class TestExecutor(job_blocks.Executor):
  """A test version of Executor with abstract methods implemented."""

  Spec = job_blocks.ExecutorSpec  # pylint: disable=invalid-name


@attr.s(auto_attribs=True)
class TestConstraint(job_blocks.Constraint):
  id: str



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions to authenticate with GCP."""

import functools
import os
from typing import Any, Iterable, Dict

from absl import flags
from google import auth
from googleapiclient import discovery
from googleapiclient import errors

_DEFAULT_SCOPES = ('https://www.googleapis.com/auth/cloud-platform',)


def get_project_name() -> str:
  """Gets the Project ID of the GCP Project."""
  _, project = auth.default()
  return project


def get_project_number() -> str:
  """Gets the Project Number of the GCP Project."""
  rm = discovery.build('cloudresourcemanager', 'v1')
  response = rm.projects().get(projectId=get_project_name()).execute()
  return response['projectNumber']


def get_creds(scopes: Iterable[str] = _DEFAULT_SCOPES):
  """Gets the google credentials to be used with GCP APIs."""
  creds, _ = auth.default(scopes=scopes)
  return creds


# The @lru_cache decorator causes this to only be run once per Python session.
@functools.lru_cache()
def enable_apis():
  """Enables APIs on the GCP Project."""
  resource = discovery.build('serviceusage', 'v1')
  body = {
      'serviceIds': [
          'aiplatform.googleapis.com',
          'cloudbuild.googleapis.com',
          'cloudresourcemanager.googleapis.com',
          'compute.googleapis.com',
          'container.googleapis.com',
          'containerregistry.googleapis.com',
          'iam.googleapis.com',
          'logging.googleapis.com',
          'storage-api.googleapis.com',
          'tpu.googleapis.com',
      ]
  }
  resource.services().batchEnable(
      parent=f'projects/{get_project_number()}', body=body).execute()


def get_service_account() -> str:
  """Gets or creates the service account for running XManager in GCP.

  The default Vertex AI Training Service Agent has limited scopes. It is more
  useful to use a custom service account that can access a greater number of
  GCP APIs.

  Returns:
    The service account email.
  Raises:
    HttpError: if the response was not a 2xx or 403.
  """
  service_account = f'xmanager@{get_project_name()}.iam.gserviceaccount.com'
  try:
    _maybe_create_service_account(service_account)
    _maybe_grant_service_account_permissions(service_account)
  except errors.HttpError as e:
    # A 403 implies that the user is not an IAM Admin.
    # The project admin probably already has set up IAM roles, so this check
    # can be skipped.
    if e.resp.status != 403:
      raise e
  return service_account


def _maybe_create_service_account(service_account: str) -> None:
  """Creates the default service account if it does not exist."""
  iam = discovery.build('iam', 'v1')
  accounts = iam.projects().serviceAccounts().list(
      name='projects/' + get_project_name()).execute()
  for account in accounts['accounts']:
    if account['email'] == service_account:
      return

  # https://cloud.google.com/iam/docs/reference/rest/v1/projects.serviceAccounts
  body = {
      'accountId': service_account.split('@')[0],
      'serviceAccount': {
          'displayName': service_account.split('@')[0],
          'description': 'XManager service account',
      },
  }
  accounts = iam.projects().serviceAccounts().create(
      name='projects/' + get_project_name(), body=body).execute()


def _maybe_grant_service_account_permissions(service_account: str) -> None:
  """Grants the default service account IAM permissions if necessary."""
  rm = discovery.build('cloudresourcemanager', 'v1')
  policy = rm.projects().getIamPolicy(resource=get_project_name()).execute()
  want_roles = ['roles/aiplatform.user', 'roles/storage.admin']

  should_set = False
  for role in want_roles:
    member = 'serviceAccount:' + service_account
    changed = _add_member_to_iam_policy(policy, role, member)
    should_set = should_set or changed

  if not should_set:
    return None

  body = {'policy': policy}
  rm.projects().setIamPolicy(resource=get_project_name(), body=body).execute()


def _add_member_to_iam_policy(policy: Dict[str, Any], role: str,
                              member: str) -> bool:
  """Modifies the IAM policy to add the member with the role."""
  for i, binding in enumerate(policy['bindings']):
    if binding['role'] == role:
      if member in binding['members']:
        return False
      policy['bindings'][i]['members'].append(member)
      return True

  policy['bindings'].append({'role': role, 'members': [member]})
  return True


def get_bucket() -> str:
  bucket = os.environ.get('GOOGLE_CLOUD_BUCKET_NAME', None)
  if bucket:
    return bucket
  raise ValueError(
      '$GOOGLE_CLOUD_BUCKET_NAME is undefined. Run '
      '`export GOOGLE_CLOUD_BUCKET_NAME=<bucket-name>`, '
      'replacing <bucket-name> with a Google Cloud Storage bucket. '
      'You can create a bucket with '
      '`gsutil mb -l us-central1 gs://$GOOGLE_CLOUD_BUCKET_NAME`')



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import threading
import unittest

from xmanager.xm import core
from xmanager.xm import job_blocks
from xmanager.xm import testing
from xmanager.xm import utils


class TestError(RuntimeError):
  """Exception which can be used in tests below."""


async def failing_job_generator(work_unit: core.WorkUnit):
  raise TestError


class ApplyArgsTest(unittest.TestCase):

  def test_wrong_job_args(self):
    with self.assertRaises(ValueError):
      core._apply_args(
          job_blocks.Job(
              job_blocks.Executable(name=''), testing.TestExecutor()),
          {'abra': 'kadabra'})

  def test_wrong_job_group_args(self):
    with self.assertRaises(ValueError):
      core._apply_args(
          job_blocks.JobGroup(
              learner=job_blocks.Job(
                  job_blocks.Executable(name=''), testing.TestExecutor())),
          {'eval': {
              'args': {
                  'batch_size': 32
              }
          }})


class ExperimentTest(unittest.TestCase):

  def test_single_job_launch(self):
    experiment = testing.TestExperiment()
    with experiment:
      job = job_blocks.Job(
          testing.TestExecutable(),
          testing.TestExecutor(),
          args={},
          name='name')
      experiment.add(job)

    self.assertEqual(experiment.launched_jobs, [job])

  def test_job_group_launch(self):
    experiment = testing.TestExperiment()
    with experiment:
      foo_job = job_blocks.Job(
          testing.TestExecutable(),
          testing.TestExecutor(),
          args={'foo': 1},
          name='1')
      bar_job = job_blocks.Job(
          testing.TestExecutable(),
          testing.TestExecutor(),
          args={'bar': 2},
          name='2')
      experiment.add(job_blocks.JobGroup(foo=foo_job, bar=bar_job))

    self.assertEqual(experiment.launched_jobs, [foo_job, bar_job])

  def test_job_generator_launch(self):
    experiment = testing.TestExperiment()
    with experiment:
      job = job_blocks.Job(
          testing.TestExecutable(),
          testing.TestExecutor(),
          args={},
          name='name')

      async def job_generator(work_unit: core.WorkUnit, use_magic: bool):
        self.assertEqual(use_magic, True)
        work_unit.add(job)

      experiment.add(job_generator, args={'use_magic': True})

    self.assertEqual(experiment.launched_jobs, [job])
    self.assertEqual(experiment.launched_jobs_args, [{'use_magic': True}])

  def test_job_generator_raises(self):
    experiment = testing.TestExperiment()
    with self.assertRaises(TestError):
      with experiment:
        experiment.add(failing_job_generator)

  def test_non_async_job_generator_raises_user_friendly_exception(self):
    with self.assertRaisesRegex(ValueError, '.* generator must be an async .*'):
      with testing.TestExperiment() as experiment:

        def job_generator(work_unit: core.WorkUnit):
          del work_unit

        experiment.add(job_generator)

  def test_launch_with_args(self):
    experiment = testing.TestExperiment()
    with experiment:
      experiment.add(
          job_blocks.JobGroup(
              foo=job_blocks.Job(
                  testing.TestExecutable(),
                  testing.TestExecutor(),
                  args={
                      'x': 1,
                      'y': 2
                  },
                  env_vars={'EDITOR': 'vi'}),
              bar=job_blocks.Job(
                  testing.TestExecutable(),
                  testing.TestExecutor(),
                  args=['--bar=1'])),
          args={
              'foo': {
                  'args': {
                      'x': 3,
                      'z': 4
                  },
                  'env_vars': {
                      'TURBO': 'ON'
                  }
              },
              'bar': {
                  'args': ['--spacebar']
              },
          })

    self.assertEqual(
        experiment.launched_jobs[0].args,
        job_blocks.SequentialArgs.from_collection({
            'x': 3,
            'y': 2,
            'z': 4
        }),
    )
    self.assertEqual(experiment.launched_jobs[0].env_vars, {
        'TURBO': 'ON',
        'EDITOR': 'vi'
    })
    self.assertEqual(
        experiment.launched_jobs[1].args,
        job_blocks.SequentialArgs.from_collection(['--bar=1', '--spacebar']),
    )

  def test_add_runs_asynchronously(self):
    generator_called = threading.Event()

    with testing.TestExperiment() as experiment:

      async def job_generator(work_unit: core.WorkUnit):
        del work_unit
        generator_called.set()

      experiment.add(job_generator)

      # Validate that job_generator is executed in a parallel thread.
      self.assertTrue(generator_called.wait(timeout=5))

  @utils.run_in_asyncio_loop
  async def test_loop_is_reused_in_coro_context(self):
    loop = asyncio.get_event_loop()

    async with testing.TestExperiment() as experiment:

      async def job_generator(work_unit: core.WorkUnit):
        del work_unit
        self.assertEqual(asyncio.get_event_loop(), loop)

      experiment.add(job_generator)

  @utils.run_in_asyncio_loop
  async def test_sync_with_cant_be_used_in_coro_context(self):
    # `async with` works.
    async with testing.TestExperiment():
      pass

    with self.assertRaises(RuntimeError):
      # But `with` raises an exception.
      with testing.TestExperiment():
        pass

  @utils.run_in_asyncio_loop
  async def test_work_unit_wait_until_complete(self):
    experiment = testing.TestExperiment()
    async with experiment:
      experiment.add(
          job_blocks.Job(
              testing.TestExecutable(), testing.TestExecutor(), args={}))
      await experiment.work_units[0].wait_until_complete()

  @utils.run_in_asyncio_loop
  async def test_work_unit_wait_until_complete_exception(self):
    experiment = testing.TestExperiment()
    with self.assertRaises(TestError):
      async with experiment:
        experiment.add(failing_job_generator)
        with self.assertRaises(core.ExperimentUnitError):
          await experiment.work_units[0].wait_until_complete()

  @utils.run_in_asyncio_loop
  async def test_get_full_job_name(self):

    async def generator(work_unit):
      self.assertEqual(work_unit.get_full_job_name('name'), '1_1_name')

    async with testing.TestExperiment() as experiment:
      experiment.add(generator)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convenience methods for constructing core objects."""

from typing import Collection, List, Mapping, Optional, Union

import immutabledict
from xmanager.xm import executables
from xmanager.xm import job_blocks


# NOTE: Do not edit methods below manually.
# They should be generated with packagables_generator utility.


def binary(
    executor_spec: job_blocks.ExecutorSpec,
    path: str,
    dependencies: Collection[executables.BinaryDependency] = (),
    *,
    args: Optional[job_blocks.UserArgs] = None,
    env_vars: Mapping[str, str] = immutabledict.immutabledict(),
) -> job_blocks.Packageable:
  # pyformat: disable
  """A prebuilt executable program.

  Args:
    executor_spec: Where the binary should be launched. Instructs for which
      platform it should be packaged.
    path: Path to a prebuilt binary.
    dependencies: A list of data dependencies to be packaged together with the
      binary.
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as two arguments: --batch_size 16. If
      positional arguments are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to be set.

  Returns:
    A packageable object which can be turned into an executable with
    Experiment.package or Experiment.package_async.
  """
  # pyformat: enable
  return job_blocks.Packageable(
      executable_spec=executables.Binary(
          path=path,
          dependencies=dependencies,
      ),
      executor_spec=executor_spec,
      args=args,
      env_vars=env_vars,
  )


def bazel_binary(
    executor_spec: job_blocks.ExecutorSpec,
    label: str,
    dependencies: Collection[executables.BinaryDependency] = (),
    bazel_args: Collection[str] = (),
    *,
    args: Optional[job_blocks.UserArgs] = None,
    env_vars: Mapping[str, str] = immutabledict.immutabledict(),
) -> job_blocks.Packageable:
  # pyformat: disable
  """A Bazel target that produces a self-contained binary.

  Note that for Python targets based on https://github.com/google/subpar
  a self-contained '.par' binary would be built.

  Args:
    executor_spec: Where the binary should be launched. Instructs for which
      platform it should be packaged.
    label: The Bazel target to be built.
    dependencies: A list of data dependencies to be packaged together with
      the binary.
    bazel_args: Bazel command line arguments.
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as two arguments: --batch_size 16. If
      positional arguments are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to be set.

  Returns:
    A packageable object which can be turned into an executable with
    Experiment.package or Experiment.package_async.
  """
  # pyformat: enable
  return job_blocks.Packageable(
      executable_spec=executables.BazelBinary(
          label=label,
          dependencies=dependencies,
          bazel_args=bazel_args,
      ),
      executor_spec=executor_spec,
      args=args,
      env_vars=env_vars,
  )


def container(
    executor_spec: job_blocks.ExecutorSpec,
    image_path: str,
    *,
    args: Optional[job_blocks.UserArgs] = None,
    env_vars: Mapping[str, str] = immutabledict.immutabledict(),
) -> job_blocks.Packageable:
  # pyformat: disable
  """A prebuilt Docker image.

  The image can be tagged locally or in a remote repository.

  Args:
    executor_spec: Where the binary should be launched. Instructs for which
      platform it should be packaged.
    image_path: Path to a prebuilt container image.
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as two arguments: --batch_size 16. If
      positional arguments are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to be set.

  Returns:
    A packageable object which can be turned into an executable with
    Experiment.package or Experiment.package_async.
  """
  # pyformat: enable
  return job_blocks.Packageable(
      executable_spec=executables.Container(image_path=image_path,),
      executor_spec=executor_spec,
      args=args,
      env_vars=env_vars,
  )


def bazel_container(
    executor_spec: job_blocks.ExecutorSpec,
    label: str,
    bazel_args: Collection[str] = (),
    *,
    args: Optional[job_blocks.UserArgs] = None,
    env_vars: Mapping[str, str] = immutabledict.immutabledict(),
) -> job_blocks.Packageable:
  # pyformat: disable
  """A Bazel target that produces a .tar image.

  Note that for targets based on https://github.com/bazelbuild/rules_docker one
  should append '.tar' to the label to specify a self-contained image.

  Args:
    executor_spec: Where the binary should be launched. Instructs for which
      platform it should be packaged.
    label: The Bazel target to be built.
    bazel_args: Bazel command line arguments.
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as two arguments: --batch_size 16. If
      positional arguments are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to be set.

  Returns:
    A packageable object which can be turned into an executable with
    Experiment.package or Experiment.package_async.
  """
  # pyformat: enable
  return job_blocks.Packageable(
      executable_spec=executables.BazelContainer(
          label=label,
          bazel_args=bazel_args,
      ),
      executor_spec=executor_spec,
      args=args,
      env_vars=env_vars,
  )


def python_container(
    executor_spec: job_blocks.ExecutorSpec,
    entrypoint: Union[executables.ModuleName, executables.CommandList],
    path: str = '.',
    base_image: Optional[str] = None,
    docker_instructions: Optional[List[str]] = None,
    use_deep_module: bool = False,
    *,
    args: Optional[job_blocks.UserArgs] = None,
    env_vars: Mapping[str, str] = immutabledict.immutabledict(),
) -> job_blocks.Packageable:
  # pyformat: disable
  """PythonContainer describes a directory containing Python code.

  Args:
    executor_spec: Where the binary should be launched. Instructs for which
      platform it should be packaged.
    entrypoint: The Python module or list of shell commands to run when entering
      this Python project.
    path: Relative or absolute path to the Python project. By default, the
      current directory (`'.'`) is used.
    base_image: Name of the image to initialize a new Docker build stage using
      the instruction `FROM`.
    docker_instructions: List of Docker instructions to apply when building the
      image. If not specified, the default one will be provided.

      When you use `docker_instructions`, you are responsible for copying the
      project directory. For example, if you are running with:

        path='/path/to/cifar10'

      You should include these steps in your `docker_instructions`:

        [
          'COPY cifar10/ cifar10',
          'WORKDIR cifar10',
        ]

      If your source code rarely changes, you can make this your first step.
      If you are frequently iterating on the source code, it is best practice
      to place these steps as late as possible in the list to maximize Docker
      layer-caching.
    use_deep_module: Whether the experiment code uses deep module structure
      (i.e., 'from <a.prefix> import models') or not (i.e., 'import models').

      If use_deep_module is set to True, and docker_instructions are used, it
      is recommended to use dedicated workdir and copy a whole project
      directory there. The example above should be modified as:

        [
          'RUN mkdir /workdir',
          'WORKDIR /workdir',
          'COPY cifar10/ /workdir/cifar10',
        ]
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as two arguments: --batch_size 16. If
      positional arguments are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to be set.

  Returns:
    A packageable object which can be turned into an executable with
    Experiment.package or Experiment.package_async.
  """
  # pyformat: enable
  return job_blocks.Packageable(
      executable_spec=executables.PythonContainer(
          entrypoint=entrypoint,
          path=path,
          base_image=base_image,
          docker_instructions=docker_instructions,
          use_deep_module=use_deep_module,
      ),
      executor_spec=executor_spec,
      args=args,
      env_vars=env_vars,
  )


def dockerfile_container(
    executor_spec: job_blocks.ExecutorSpec,
    path: str = '.',
    dockerfile: Optional[str] = None,
    *,
    args: Optional[job_blocks.UserArgs] = None,
    env_vars: Mapping[str, str] = immutabledict.immutabledict(),
) -> job_blocks.Packageable:
  # pyformat: disable
  """Dockerfile describes a Dockerfile for generating a docker image.

  This is a lower-level feature that could be solved using higher-level
  Executables such as BazelContainer or PythonContainer.

  Args:
    executor_spec: Where the binary should be launched. Instructs for which
      platform it should be packaged.
    path: Specifies the build's context.
    dockerfile: The file that will be used for build instructions. Otherwise,
      {path}/Dockerfile will be used. Equivalent to `docker build -f`. A
      relative path will use a Dockerfile that is relative to the launcher
      script.
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as two arguments: --batch_size 16. If
      positional arguments are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to be set.

  Returns:
    A packageable object which can be turned into an executable with
    Experiment.package or Experiment.package_async.
  """
  # pyformat: enable
  return job_blocks.Packageable(
      executable_spec=executables.Dockerfile(
          path=path,
          dockerfile=dockerfile,
      ),
      executor_spec=executor_spec,
      args=args,
      env_vars=env_vars,
  )



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper module for using TPUs."""
from typing import List


# pylint: disable=line-too-long
def tpuvm_docker_instructions() -> List[str]:
  return [
      'RUN wget '
      'https://storage.googleapis.com/cloud-tpu-tpuvm-artifacts/libtpu/20210525/libtpu.so'
      ' -O /lib/libtpu.so',
      'RUN chmod 700 /lib/libtpu.so',
      'RUN wget '
      'https://storage.googleapis.com/cloud-tpu-tpuvm-artifacts/tensorflow/20210525/tf_nightly-2.6.0-cp38-cp38-linux_x86_64.whl',
      'RUN pip3 install tf_nightly-2.6.0-cp38-cp38-linux_x86_64.whl',
      'RUN rm tf_nightly-2.6.0-cp38-cp38-linux_x86_64.whl',
  ]



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""XManager client API.

Provides XManager public API for configuring and launching experiments.
"""

from xmanager.xm import job_operators
from xmanager.xm.compute_units import *
from xmanager.xm.core import *
from xmanager.xm.executables import *
from xmanager.xm.job_blocks import *
from xmanager.xm.metadata_context import ContextAnnotations
from xmanager.xm.metadata_context import MetadataContext
from xmanager.xm.packagables import *
from xmanager.xm.resources import *
from xmanager.xm.utils import run_in_asyncio_loop
from xmanager.xm.utils import ShellSafeArg



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the xmanager.contrib.gcs module."""

import getpass

from absl import app
from absl.testing import absltest
from absl.testing import flagsaver
from absl.testing import parameterized
from xmanager.contrib import gcs

# Error patterns.
_GCS_PATH_ERROR = ('--xm_gcs_path not in gs://bucket/directory or /gcs/path '
                   'format.')
_PATH_ERROR = 'Path not in gs://bucket/directory or /gcs/path format'


class GcsTest(parameterized.TestCase):

  def gcs_prefix(self, project_name: str) -> str:
    return ('gs://xcloud_public_bucket/' + getpass.getuser() + '/' +
            project_name + '-')

  def test_gcs_path_suggestions(self):
    self.assertStartsWith(gcs.suggestion('xcloud'), self.gcs_prefix('xcloud'))

  def test_gcs_path_empty_flag(self):
    self.assertEqual(
        gcs.get_gcs_path_or_suggestion('xcloud'), gcs.suggestion('xcloud'))
    with self.assertRaisesRegex(app.UsageError, '--xm_gcs_path is missing'):
      gcs.get_gcs_path_or_fail('xcloud')

  def test_gcs_path_correct_value(self):
    with flagsaver.flagsaver(xm_gcs_path='gs://bucket/dir'):
      self.assertEqual(
          gcs.get_gcs_path_or_suggestion('xcloud'), 'gs://bucket/dir')
      self.assertEqual(gcs.get_gcs_path_or_fail('xcloud'), 'gs://bucket/dir')

  def test_gcs_path_incorrect_value(self):
    with flagsaver.flagsaver(xm_gcs_path='file://dir'):
      with self.assertRaisesRegex(app.UsageError, _GCS_PATH_ERROR):
        gcs.get_gcs_path_or_suggestion('xcloud')
        gcs.get_gcs_path_or_fail('xcloud')

  # pylint: disable=bad-whitespace
  @parameterized.named_parameters(
      ('gs_long',      'gs://a/b/c', True,  False, True),
      ('gs_short',     'gs://d/e',   True,  False, True),
      ('gs_invalid',   'gs:/d/e',    False, False, False),
      ('fuse_long',    '/gcs/a/b/c', False, True,  True),
      ('fuse_short',   '/gcs/d/e',   False, True,  True),
      ('fuse_invalid', '/gcsc/d/e',  False, False, False),
      ('invalid',      'a/b/f',      False, False, False),
  )  # pyformat:disable
  # pylint: enable=bad-whitespace
  def test_is_path(self, path, expected_gs, expected_fuse, expected_gcs):
    self.assertEqual(gcs.is_gs_path(path), expected_gs)
    self.assertEqual(gcs.is_gcs_fuse_path(path), expected_fuse)
    self.assertEqual(gcs.is_gcs_path(path), expected_gcs)

  def test_get_gcs_url(self):
    self.assertEqual(
        gcs.get_gcs_url('gs://a/b/c'),
        f'{gcs.gcp_website_url}/storage/browser/a/b/c')
    self.assertEqual(
        gcs.get_gcs_url('gs://d/e'),
        f'{gcs.gcp_website_url}/storage/browser/d/e')
    self.assertEqual(
        gcs.get_gcs_url('/gcs/a/b/c'),
        f'{gcs.gcp_website_url}/storage/browser/a/b/c')
    self.assertEqual(
        gcs.get_gcs_url('/gcs/d/e'),
        f'{gcs.gcp_website_url}/storage/browser/d/e')
    with self.assertRaisesRegex(ValueError, _PATH_ERROR):
      gcs.get_gcs_url('a/b/f')

  def test_get_gcs_fuse_path(self):
    self.assertEqual(gcs.get_gcs_fuse_path('gs://a/b/c'), '/gcs/a/b/c')
    self.assertEqual(gcs.get_gcs_fuse_path('gs://d/e'), '/gcs/d/e')
    self.assertEqual(gcs.get_gcs_fuse_path('/gcs/a/b/c'), '/gcs/a/b/c')
    self.assertEqual(gcs.get_gcs_fuse_path('/gcs/d/e'), '/gcs/d/e')
    with self.assertRaisesRegex(ValueError, _PATH_ERROR):
      gcs.get_gcs_fuse_path('a/b/f')

  def test_get_gs_path(self):
    self.assertEqual(gcs.get_gs_path('gs://a/b/c'), 'gs://a/b/c')
    self.assertEqual(gcs.get_gs_path('gs://d/e'), 'gs://d/e')
    self.assertEqual(gcs.get_gs_path('/gcs/a/b/c'), 'gs://a/b/c')
    self.assertEqual(gcs.get_gs_path('/gcs/d/e'), 'gs://d/e')
    with self.assertRaisesRegex(ValueError, _PATH_ERROR):
      gcs.get_gcs_fuse_path('a/b/f')


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import unittest

from xmanager.xm import pattern_matching


class UtilsTest(unittest.TestCase):

  def testMatch_selectsAppropriateBranch(self):
    matcher = pattern_matching.match(
        pattern_matching.Case([int], lambda _: 'A'),
        pattern_matching.Case([str], lambda _: 'B'),
    )

    self.assertEqual(matcher('abc'), 'B')

  def testMatch_throwsWhenNoneFit(self):
    matcher = pattern_matching.match()

    with self.assertRaises(TypeError):
      matcher(0)

  def testMatch_earlierOptionWins(self):

    class P:
      pass

    class C(P):
      pass

    matcher = pattern_matching.match(
        pattern_matching.Case([P], lambda _: 'P'),
        pattern_matching.Case([C], lambda _: 'C'),
    )

    self.assertEqual(matcher(C()), 'P')

  def testMatch_asVisitor(self):

    def visit_int(n: int):
      return n * 2

    def visit_str(s: str):
      return len(s)

    matcher = pattern_matching.match(visit_int, visit_str)

    self.assertEqual(matcher(1), 2)
    self.assertEqual(matcher('zzz'), 3)

  def testMatch_multipleArguments(self):

    def visit_just_one(k: int):
      return k * 2

    def visit_too_many(k: int, l: int, m: int):
      return k * l * m

    def visit_just_right(k: int, l: int):
      return k * l

    matcher = pattern_matching.match(visit_just_one, visit_too_many,
                                     visit_just_right)

    self.assertEqual(matcher(2, 3), 6)

  def testMatch_parameterizedGenerics(self):
    """Test that generics parameters don't confuse isinstance.

    Note that the parameters don't participate in matching. We can't distinguish
    List[int] and List[str].
    """

    def first_element(l: List[int]) -> int:
      return l[0]

    self.assertEqual(pattern_matching.match(first_element)([1, 2]), 1)

  def testMatch_missingAnnotation(self):
    with self.assertRaises(ValueError):
      pattern_matching.match(lambda x: x)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup configuration specifying XManager dependencies."""

from setuptools import find_namespace_packages
from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as fh:
  long_description = fh.read()

setup(
    name='xmanager',
    version='0.1.3',
    description='A framework for managing machine learning experiments',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='DeepMind Technologies Limited',
    packages=find_namespace_packages(exclude=['examples.*']),
    package_data={'': ['*.sh', '*.sql']},
    python_requires='>=3.7',
    install_requires=[
        'absl-py',
        'async_generator',
        'attrs',
        'docker',
        'google-api-core',
        'google-api-python-client',
        'google-auth',
        'google-cloud-aiplatform>=1.4.0',
        'google-cloud-storage',
        'humanize',
        'immutabledict',
        'kubernetes',
        'sqlalchemy==1.2.19',
        'termcolor',
    ],
    entry_points={
        'console_scripts': ['xmanager = xmanager.cli.cli:entrypoint',],
    },
    # https://github.com/pypa/warehouse/blob/de4a2e5e2ec26d01bf7813da427ebc4725dccde9/warehouse/templates/packaging/detail.html#L20-L60
    project_urls={
        'Homepage': 'https://github.com/deepmind/xmanager',
        'Issue tracker': 'https://github.com/deepmind/xmanager/issues',
    },
)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Xmanager command-line interface."""

import errno
import importlib
import os
import sys

from absl import app

_DEFAULT_ZONE = 'us-west1-b'
_DEFAULT_CLUSTER_NAME = 'xmanager-via-caliban'


def main(argv):
  if len(argv) < 3:
    raise app.UsageError('There must be at least 2 command-line arguments')
  cmd = argv[1]
  if cmd == 'launch':
    launch_script = argv[2]
    if not os.path.exists(launch_script):
      raise OSError(errno.ENOENT, f'File not found: {launch_script}')
    sys.path.insert(0, os.path.abspath(os.path.dirname(launch_script)))
    launch_module, _ = os.path.splitext(os.path.basename(launch_script))
    m = importlib.import_module(launch_module)
    sys.path.pop(0)
    argv = [launch_script, '--xm_launch_script={}'.format(launch_script)
           ] + argv[3:]
    app.run(m.main, argv=argv)
  elif cmd == 'cluster':
    caliban_gke = importlib.import_module('caliban.platform.gke.cli')
    caliban_gke_types = importlib.import_module('caliban.platform.gke.types')
    subcmd = argv[2]
    args = {
        'dry_run': False,
        'cluster_name': _DEFAULT_CLUSTER_NAME,
        'zone': _DEFAULT_ZONE,
        'release_channel': caliban_gke_types.ReleaseChannel.REGULAR,
        'single_zone': True,
    }
    if subcmd == 'create':
      caliban_gke._cluster_create(args)  # pylint: disable=protected-access
    elif subcmd == 'delete':
      caliban_gke._cluster_delete(args)  # pylint: disable=protected-access
    else:
      raise app.UsageError(
          f'Subcommand `{cmd} {subcmd}` is not a supported subcommand')
  else:
    raise app.UsageError(f'Command `{cmd}` is not a supported command')


def entrypoint():
  app.run(main)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for transforming source code using copybara.

XManager primarily uses Copybara to run folder-to-folder workflows in the form:

core.workflow(
    name = "folder_to_folder",
    origin = folder.origin(),
    destination = folder.destination(),
    ...
)

Copybara allows for iterative local development on multiple platforms without
needing to add/remove platform-specific code modifications. This allows you to
preprocess source code so that it can be run on different platforms with
different executors. e.g.

local_version = run_workflow(config, 'local', path)
caip_version = run_workflow(config, 'caip', path)

local_spec = xm.PythonContainer(path=local_version, **kwargs)
caip_spec = xm.PythonContainer(path=caip_version, **kwargs)

[local_executable, caip_executable] = experiment.package([
  xm.Packageable(
    executable_spec=spec,
    executor_spec=xm_local.Local.Spec()),
  xm.Packageable(
    executable_spec=spec,
    executor_spec=xm_local.Caip.Spec())])

Copybara has no release process, so you must compile copybara yourself:
https://github.com/google/copybara
"""
import os
import subprocess
import tempfile
from typing import Optional

# Set with the compiled path to copybara e.g.
# COPYBARA_BIN = 'bazel-bin/java/com/google/copybara/copybara_deploy.jar'
COPYBARA_BIN = 'copybara'


def run_workflow(config: str,
                 workflow: str,
                 origin_folder: str,
                 destination_folder: Optional[str] = None,
                 config_root: Optional[str] = None) -> str:
  """Run a workflow in a copybara config to transform origin to destination.

  Args:
    config: Path to the Copybara config.
    workflow: Name of a workflow in copybara config.
    origin_folder: The origin folder to use as input. This will passed to
      Copybara via the source_ref argument.
    destination_folder: The destination folder to output.
    config_root: Configuration root path to be used for resolving absolute
      config labels like '//foo/bar'.

  Returns:
    The output destination folder.
  """
  origin_folder = os.path.abspath(origin_folder)
  if not destination_folder:
    destination_folder = tempfile.mkdtemp()
  command = [
      COPYBARA_BIN, config, workflow, '--ignore-noop', origin_folder,
      '--folder-dir=' + destination_folder
  ]
  if config_root:
    command += ['--config-root=' + config_root]
  print('Copybara command: ', command)
  subprocess.run(command, check=True)
  return destination_folder



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for interacting with Cloud Build."""
import datetime
import tarfile
import tempfile
import time
from typing import Any, Dict, Optional

from absl import flags
from docker.utils import utils as docker_utils
from google.cloud import storage
from googleapiclient import discovery
import termcolor

from xmanager.cloud import auth

_CLOUD_BUILD_TIMEOUT_SECONDS = flags.DEFINE_integer(
    'xm_cloud_build_timeout_seconds', 1200,
    'The amount of time that builds should be allowed to run, '
    'to second granularity.')
_USE_CLOUD_BUILD_CACHE = flags.DEFINE_boolean(
    'xm_use_cloud_build_cache',
    False,
    'Use Cloud Build cache to speed up the Docker build. '
    'An image with the same name tagged as :latest should exist.'
    'More details at https://cloud.google.com/cloud-build/docs/speeding-up-builds#using_a_cached_docker_image'  # pylint:disable=g-line-too-long
)

_USE_KANIKO = flags.DEFINE_boolean(
    'xm_use_kaniko', True,
    'Use kaniko backend for Cloud Build and enable caching.')
_KANIKO_CACHE_TTL = flags.DEFINE_string('xm_kaniko_cache_ttl', '336h',
                                        'Cache ttl to use for kaniko builds.')


class Client:
  """Cloud Build Client."""

  def __init__(self,
               project: Optional[str] = None,
               bucket: Optional[str] = None,
               credentials=None,
               cloud_build_timeout_seconds: Optional[int] = None,
               use_cloud_build_cache: Optional[bool] = None,
               use_kaniko: Optional[bool] = None,
               kaniko_cache_ttl: Optional[str] = None):
    """Create the Cloud Build Client.

    Args:
      project: Name of the GCP project to use for Cloud Build calls and for
        storing the data passed to Cloud Build. If not specified the project of
        the default credentials for the current environment is used.
      bucket: Bucket used to store data passed to Cloud Build. If not specified
        uses the value from the GOOGLE_CLOUD_BUCKET_NAME environment variable.
      credentials: OAuth2 Credentials to use for Cloud Build & storage calls. If
        None gets the default credentials for the current environment.
      cloud_build_timeout_seconds: The amount of time that builds should be
        allowed to run. If None defaults to `--xm_cloud_build_timeout_seconds.
      use_cloud_build_cache: Whether to use Cloud Build cache to speed up the
        Docker build. If None defaults to `--xm_use_cloud_build_cache`. An image
        with the same name tagged as :latest should exist. More details at
          https://cloud.google.com/cloud-build/docs/speeding-up-builds#using_a_cached_docker_image
      use_kaniko: Use kaniko backend for Cloud Build and enable caching. If None
        defaults to `--xm_use_kaniko`.
      kaniko_cache_ttl: Cache ttl to use for kaniko builds. If None defaults to
        `--xm_kaniko_cache_ttl`.
    """
    self.project = project or auth.get_project_name()
    self.bucket = bucket or auth.get_bucket()
    self.credentials = credentials or auth.get_creds()
    if cloud_build_timeout_seconds is None:
      cloud_build_timeout_seconds = _CLOUD_BUILD_TIMEOUT_SECONDS.value
    self.cloud_build_timeout_seconds = cloud_build_timeout_seconds
    if use_cloud_build_cache is None:
      use_cloud_build_cache = _USE_CLOUD_BUILD_CACHE.value
    self.use_cloud_build_cache = use_cloud_build_cache
    if use_kaniko is None:
      use_kaniko = _USE_KANIKO.value
    self.use_kaniko = use_kaniko
    if kaniko_cache_ttl is None:
      kaniko_cache_ttl = _KANIKO_CACHE_TTL.value
    self.kaniko_cache_ttl = kaniko_cache_ttl
    self.cloudbuild_api = None  # discovery CloudBuild v1 client

  def upload_tar_to_storage(self, archive_path: str, destination_name: str):
    storage_client = storage.Client(
        project=self.project, credentials=self.credentials)
    bucket = storage_client.bucket(self.bucket)
    blob = bucket.blob(destination_name)
    blob.upload_from_filename(archive_path)

  def build_docker_image(self, image: str, directory: str,
                         upload_name: str) -> str:
    """Create a Docker image via Cloud Build and push to Cloud Repository."""
    repository, tag = docker_utils.parse_repository_tag(image)
    if not tag:
      tag = datetime.datetime.now().strftime('%Y%m%d-%H%M%S_%f')

    _, archive_path = tempfile.mkstemp(suffix='.tar.gz')
    with tarfile.open(archive_path, 'w:gz') as tar:
      tar.add(directory, '/')
    destination_name = f'{upload_name}-{tag}.tar.gz'
    self.upload_tar_to_storage(archive_path, destination_name)
    build_body = self._build_request_body(destination_name, repository, tag)
    # Note: On GCP cache_discovery=True (the default) leads to ugly error
    # messages as file_cache is unavailable.
    if not self.cloudbuild_api:
      self.cloudbuild_api = discovery.build(
          'cloudbuild',
          'v1',
          credentials=self.credentials,
          cache_discovery=False)
    create_op = self.cloudbuild_api.projects().builds().create(
        projectId=self.project, body=build_body).execute()
    log_url = create_op['metadata']['build']['logUrl']
    print('Cloud Build link:', termcolor.colored(log_url, color='blue'))

    build_id = create_op['metadata']['build']['id']
    return self.wait_for_build(build_id, f'{repository}:{tag}')

  def _build_request_body(self, bucket_path, repository, tag) -> Dict[str, Any]:
    """Builds the Cloud Build create_build_request body."""
    body = {
        'source': {
            'storageSource': {
                'bucket': self.bucket,
                'object': bucket_path,
            },
        },
        'timeout': str(self.cloud_build_timeout_seconds) + 's',
    }
    if self.use_kaniko:
      body.update({
          'steps': [{
              'name':
                  'gcr.io/kaniko-project/executor:latest',
              'args': [
                  f'--destination={repository}:{tag}',
                  f'--destination={repository}:latest', '--cache=true',
                  f'--cache-ttl={self.kaniko_cache_ttl}'
              ],
          }]
      })
    else:
      args_for_cached_image = (['--cache-from', f'{repository}:latest']
                               if self.use_cloud_build_cache else [])
      body.update({
          'steps': [{
              'name':
                  'gcr.io/cloud-builders/docker',
              'args': [
                  'build', '-t', f'{repository}:{tag}', '-t',
                  f'{repository}:latest'
              ] + args_for_cached_image + ['.'],
          }],
          'options': {
              'machineType': 'E2_HIGHCPU_32'
          },
          'images': [repository]
      })
    return body

  def wait_for_build(self, build_id: str, kaniko_image: str) -> str:
    """Waits for build to finish and return the image URI of the result."""
    backoff = 30  # seconds
    while True:
      time.sleep(backoff)
      result = self.cloudbuild_api.projects().builds().get(
          projectId=self.project, id=build_id).execute()
      status = result['status']
      print('Cloud Build status:', status)

      if status == 'SUCCESS':
        image_uri = kaniko_image
        if not self.use_kaniko:
          # Note: Not sure if this is needed. Could we always use the uri above?
          image = result['results']['images'][0]
          image_uri = f'{image["name"]}@{image["digest"]}'
        break
      elif status == 'FAILURE':
        print('Build FAILED. See logs for more information:',
              termcolor.colored(result['logUrl'], color='red'))
        raise RuntimeError('Build FAILED.')
      elif status == 'QUEUED' or status == 'WORKING':
        continue
      elif status == 'INTERNAL_ERROR' or status == 'CANCELLED':
        print('Cloud Build tool failure. Status:', status)
        raise RuntimeError('Cloud Build tool failed. Try again.')
      else:
        print('Build not complete. See logs for more information:',
              termcolor.colored(result['logUrl'], color='red'))
        raise RuntimeError('Build FAILED.')

    print('Your image URI is:', termcolor.colored(image_uri, color='blue'))
    print('You can run your image locally via:\n' +
          termcolor.colored('docker run ' + image_uri, color='green'))
    return image_uri



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Vizier Controller."""

import time
from typing import Any, Callable, Dict, Optional

from google.cloud import aiplatform_v1beta1 as aip

from xmanager import xm


class VizierController:
  """A Controller that runs Vizier suggested hyperparameters in multiple work units."""

  def __init__(self, work_unit_generator: Callable[[Dict[str, Any]],
                                                   xm.WorkUnit],
               vz_client: aip.VizierServiceClient, study: aip.Study,
               num_work_units_total: int, num_parallel_work_units: int) -> None:
    """Create a VizierController.

    Args:
      work_unit_generator: the function that generates WorkUnit from
        hyperparameters.
      vz_client: the Vizier Client used for interacting with Vizier.
      study: the study the controller works on.
      num_work_units_total: number of work units to create in total. (TODO:
        remove this and retrieve from study spec stopping criteria once it is
        settable there.)
      num_parallel_work_units: number of work units to run in parallel.
    """
    self._work_unit_generator = work_unit_generator
    self._vz_client = vz_client
    self._study = study
    self._num_work_units_total = num_work_units_total
    self._num_parallel_work_units = num_parallel_work_units

    self._work_unit_updaters = []

  def run(self, poll_frequency_in_sec: float = 60) -> None:
    """Peridically check and sync status between vizier and work units and create new work units when needed."""
    while True:
      # 1. Complete trial for completed work unit; Early stop first if needed.
      for work_unit_updater in self._work_unit_updaters:
        if not work_unit_updater.completed:
          work_unit_updater.check_for_completion()

      # 2. TODO: Return by Vizier's indication that study is done
      # when such API is ready on Vizier side.
      num_exisiting_work_units = len(self._work_unit_updaters)
      num_completed_work_units = sum(
          [wuu.completed for wuu in self._work_unit_updaters])
      if num_exisiting_work_units == self._num_work_units_total and num_completed_work_units == self._num_work_units_total:
        print('All done! Exiting VizierController... \n')
        return

      # 3. Get new trials and assign to new work units.
      self._launch_new_work_units()

      time.sleep(poll_frequency_in_sec)

  def _launch_new_work_units(self) -> None:
    """Get hyperparmeter suggestions from Vizier and assign to new work units to run."""
    # 1. Compute num of work units to create next.
    num_existing_work_units = len(self._work_unit_updaters)
    num_running_work_units = len([
        wuu for wuu in self._work_unit_updaters
        if wuu.work_unit_status().is_active
    ])
    num_work_units_to_create_total = self._num_work_units_total - num_existing_work_units
    num_work_units_to_create_next = min(
        self._num_parallel_work_units - num_running_work_units,
        num_work_units_to_create_total)

    # 2. Create the work units.
    start_index = num_existing_work_units + 1
    for i in range(start_index, start_index + num_work_units_to_create_next):
      trial = self._vz_client.suggest_trials(
          request=aip.SuggestTrialsRequest(
              parent=self._study.name,
              suggestion_count=1,
              client_id=f'work unit {i}')).result().trials[0]
      print(f'Trial for work unit (index: {i}) is retrieved：\n{trial}')

      print(f'Creating work unit (index: {i})... \n')

      work_unit = self._work_unit_generator({
          'trial_name': trial.name,
          **{
              p.parameter_id: getattr(p.value, p.value.WhichOneof('kind'))
              for p in trial.parameters
          }
      })

      # TODO: Add an utility to handle logging conditionally
      # (use print when run local otherwise logging.info.)
      print(f'Work unit (index: {i}, id: {work_unit.work_unit_id}) created. \n')

      self._work_unit_updaters.append(
          WorkUnitVizierUpdater(
              vz_client=self._vz_client, work_unit=work_unit, trial=trial))


class WorkUnitVizierUpdater:
  """An updater for syncing completion state between work unit and vizier trial."""

  def __init__(self, vz_client: aip.VizierServiceClient, work_unit: xm.WorkUnit,
               trial: aip.Trial) -> None:
    self.completed = False
    self._vz_client = vz_client
    self._work_unit = work_unit
    self._trial = trial

  def work_unit_status(self) -> xm.ExperimentUnitStatus:
    return self._work_unit.get_status()

  def check_for_completion(self) -> None:
    """Sync the completion status between WorkUnit and Vizier Trial if needed."""
    if self.completed:
      return

    print(
        f'Start completion check for work unit {self._work_unit.work_unit_id}.\n'
    )

    # TODO: Add infeasible_reason when available.
    if not self.work_unit_status().is_active:
      self._complete_trial(self._trial)
      self.completed = True
    elif self._vz_client.check_trial_early_stopping_state(
        request=aip.CheckTrialEarlyStoppingStateRequest(
            trial_name=self._trial.name)).result().should_stop:
      print(f'Early stopping work unit {self._work_unit.work_unit_id}.\n')
      self._work_unit.stop()
    else:
      print(f'Work unit {self._work_unit.work_unit_id} is still running.\n')

  def _complete_trial(self,
                      trial: aip.Trial,
                      infeasible_reason: Optional[str] = None) -> None:
    """Complete a trial."""
    self._vz_client.complete_trial(
        request=aip.CompleteTrialRequest(
            name=trial.name,
            trial_infeasible=infeasible_reason is not None,
            infeasible_reason=infeasible_reason))
    print(f'Trial {trial.name} is completed\n')



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Local backend executables."""

from typing import Dict

import attr
from xmanager import xm


@attr.s(auto_attribs=True)
class LoadedContainerImage(xm.Executable):
  """A locally loaded container image."""

  image_id: str
  args: xm.SequentialArgs = attr.Factory(xm.SequentialArgs)
  env_vars: Dict[str, str] = attr.Factory(dict)


@attr.s(auto_attribs=True)
class LocalBinary(xm.Executable):
  """A locally located binary."""

  path: str
  args: xm.SequentialArgs = attr.Factory(xm.SequentialArgs)
  env_vars: Dict[str, str] = attr.Factory(dict)


@attr.s(auto_attribs=True)
class GoogleContainerRegistryImage(xm.Executable):
  """An image inside Google Container Registry."""

  image_path: str
  args: xm.SequentialArgs = attr.Factory(xm.SequentialArgs)
  env_vars: Dict[str, str] = attr.Factory(dict)



# Copyright 2017 The Bazel Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyformat: disable
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# pylint: skip-file
# source: src/main/protobuf/option_filters.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/main/protobuf/option_filters.proto',
  package='options',
  syntax='proto3',
  serialized_options=b'\n(com.google.devtools.common.options.proto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&src/main/protobuf/option_filters.proto\x12\x07options*\xea\x02\n\x0fOptionEffectTag\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05NO_OP\x10\x01\x12\x1b\n\x17LOSES_INCREMENTAL_STATE\x10\x02\x12\x12\n\x0e\x43HANGES_INPUTS\x10\x03\x12\x13\n\x0f\x41\x46\x46\x45\x43TS_OUTPUTS\x10\x04\x12\x18\n\x14\x42UILD_FILE_SEMANTICS\x10\x05\x12 \n\x1c\x42\x41ZEL_INTERNAL_CONFIGURATION\x10\x06\x12\x18\n\x14LOADING_AND_ANALYSIS\x10\x07\x12\r\n\tEXECUTION\x10\x08\x12\'\n#HOST_MACHINE_RESOURCE_OPTIMIZATIONS\x10\t\x12\x15\n\x11\x45\x41GERNESS_TO_EXIT\x10\n\x12\x14\n\x10\x42\x41ZEL_MONITORING\x10\x0b\x12\x13\n\x0fTERMINAL_OUTPUT\x10\x0c\x12\x18\n\x14\x41\x43TION_COMMAND_LINES\x10\r\x12\x0f\n\x0bTEST_RUNNER\x10\x0e*\x95\x01\n\x11OptionMetadataTag\x12\x10\n\x0c\x45XPERIMENTAL\x10\x00\x12\x17\n\x13INCOMPATIBLE_CHANGE\x10\x01\x12\x0e\n\nDEPRECATED\x10\x02\x12\n\n\x06HIDDEN\x10\x03\x12\x0c\n\x08INTERNAL\x10\x04\"\x04\x08\x05\x10\x05*%TRIGGERED_BY_ALL_INCOMPATIBLE_CHANGESB*\n(com.google.devtools.common.options.protob\x06proto3'
)

_OPTIONEFFECTTAG = _descriptor.EnumDescriptor(
  name='OptionEffectTag',
  full_name='options.OptionEffectTag',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NO_OP', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOSES_INCREMENTAL_STATE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CHANGES_INPUTS', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AFFECTS_OUTPUTS', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BUILD_FILE_SEMANTICS', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BAZEL_INTERNAL_CONFIGURATION', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOADING_AND_ANALYSIS', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EXECUTION', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HOST_MACHINE_RESOURCE_OPTIMIZATIONS', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EAGERNESS_TO_EXIT', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BAZEL_MONITORING', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TERMINAL_OUTPUT', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTION_COMMAND_LINES', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEST_RUNNER', index=14, number=14,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=52,
  serialized_end=414,
)
_sym_db.RegisterEnumDescriptor(_OPTIONEFFECTTAG)

OptionEffectTag = enum_type_wrapper.EnumTypeWrapper(_OPTIONEFFECTTAG)
_OPTIONMETADATATAG = _descriptor.EnumDescriptor(
  name='OptionMetadataTag',
  full_name='options.OptionMetadataTag',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EXPERIMENTAL', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCOMPATIBLE_CHANGE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DEPRECATED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HIDDEN', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=417,
  serialized_end=566,
)
_sym_db.RegisterEnumDescriptor(_OPTIONMETADATATAG)

OptionMetadataTag = enum_type_wrapper.EnumTypeWrapper(_OPTIONMETADATATAG)
UNKNOWN = 0
NO_OP = 1
LOSES_INCREMENTAL_STATE = 2
CHANGES_INPUTS = 3
AFFECTS_OUTPUTS = 4
BUILD_FILE_SEMANTICS = 5
BAZEL_INTERNAL_CONFIGURATION = 6
LOADING_AND_ANALYSIS = 7
EXECUTION = 8
HOST_MACHINE_RESOURCE_OPTIMIZATIONS = 9
EAGERNESS_TO_EXIT = 10
BAZEL_MONITORING = 11
TERMINAL_OUTPUT = 12
ACTION_COMMAND_LINES = 13
TEST_RUNNER = 14
EXPERIMENTAL = 0
INCOMPATIBLE_CHANGE = 1
DEPRECATED = 2
HIDDEN = 3
INTERNAL = 4


DESCRIPTOR.enum_types_by_name['OptionEffectTag'] = _OPTIONEFFECTTAG
DESCRIPTOR.enum_types_by_name['OptionMetadataTag'] = _OPTIONMETADATATAG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for building Docker images."""
import datetime
import os
import pathlib
import shutil
import subprocess
import sys
from typing import Optional

from absl import logging
import docker
from docker.utils import utils as docker_utils
import humanize
import termcolor


def create_tag() -> str:
  return datetime.datetime.now().strftime('%Y%m%d-%H%M%S-%f')


def prepare_directory(destination_directory: str, source_directory: str,
                      project_name: str, entrypoint_file: str,
                      dockerfile: str) -> None:
  """Stage all inputs into the destination directory.

  Args:
    destination_directory: The directory to copy files to.
    source_directory: The directory to copy files from.
    project_name: The name of the folder inside destination_directory/ that
      source_directory/ files will be copied to.
    entrypoint_file: The file path of entrypoint.sh.
    dockerfile: The file path of Dockerfile.
  """
  source_path = pathlib.Path(source_directory)
  size = sum(f.stat().st_size for f in source_path.glob('**/*') if f.is_file())
  print(f'Size of Docker input: {humanize.naturalsize(size)}')
  if size > 200 * 10**6:
    print(
        termcolor.colored(
            'You are trying to pack over 200MB into a Docker image. '
            'Large images negatively impact build times',
            color='magenta'))
  shutil.copytree(source_directory,
                  os.path.join(destination_directory, project_name))
  shutil.copyfile(dockerfile, os.path.join(destination_directory, 'Dockerfile'))
  shutil.copyfile(entrypoint_file,
                  os.path.join(destination_directory, 'entrypoint.sh'))


def build_docker_image(image: str,
                       directory: str,
                       dockerfile: Optional[str] = None,
                       use_docker_command: bool = True,
                       show_docker_command_progress: bool = False) -> str:
  """Builds a Docker image locally."""
  logging.info('Building Docker image')
  docker_client = docker.from_env()
  if not dockerfile:
    dockerfile = os.path.join(directory, 'Dockerfile')
  if use_docker_command:
    _build_image_with_docker_command(docker_client, directory, image,
                                     dockerfile, show_docker_command_progress)
  else:
    _build_image_with_python_client(docker_client, directory, image, dockerfile)
  logging.info('Building docker image: Done')
  return image


def push_docker_image(image: str) -> str:
  """Pushes a Docker image to the designated repository."""
  docker_client = docker.from_env()
  repository, tag = docker_utils.parse_repository_tag(image)
  push = docker_client.images.push(repository=repository, tag=tag)
  logging.info(push)
  if not isinstance(push, str) or '"Digest":' not in push:
    raise RuntimeError(
        'Expected docker push to return a string with `status: Pushed` and a '
        'Digest. This is probably a temporary issue with --build_locally and '
        'you should try again')
  # If we are pushing an image, then :latest should also be present.
  docker_client.images.push(repository=repository, tag='latest')
  print('Your image URI is:', termcolor.colored(image, color='blue'))
  return image


def _build_image_with_docker_command(client: docker.DockerClient,
                                     path: str,
                                     image_tag: str,
                                     dockerfile: str,
                                     progress: bool = False) -> None:
  """Builds a Docker image by calling `docker build` within a subprocess."""
  version = client.version()['Version']
  [major, minor] = version.split('.')[:2]
  if float(f'{major}.{minor}') < 20.10:
    # docker buildx requires docker 20.10.
    raise RuntimeError('XCloud requires Docker Engine version 20.10+.')
  repository, tag = docker_utils.parse_repository_tag(image_tag)
  if not tag:
    tag = 'latest'
  command = [
      'docker', 'buildx', 'build', '-t', f'{repository}:{tag}', '-t',
      f'{repository}:latest', '-f', dockerfile, path
  ]

  # Adding flags to show progress and disabling cache.
  # Caching prevents actual commands in layer from executing.
  # This is turn makes displaying progress redundant.
  if progress:
    command[2:2] = ['--progress', 'plain', '--no-cache']

  subprocess.run(command, check=True, env={'DOCKER_BUILDKIT': '1'})


def _build_image_with_python_client(client: docker.DockerClient, path: str,
                                    image_tag: str, dockerfile: str) -> None:
  """Builds a Docker image by calling the Docker Python client."""
  repository, tag = docker_utils.parse_repository_tag(image_tag)
  if not tag:
    tag = 'latest'
  try:
    # The `tag=` arg refers to the full repository:tag image name.
    _, logs = client.images.build(
        path=path, tag=f'{repository}:{tag}', dockerfile=dockerfile)
    client.images.build(
        path=path, tag=f'{repository}:latest', dockerfile=dockerfile)
  except docker.errors.BuildError as error:
    for log in error.build_log:
      print(log.get('stream', ''), end='', file=sys.stderr)
    raise error
  for log in logs:
    print(log.get('stream', ''), end='')



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for interacting with Kubernetes."""

import asyncio
import functools
from typing import Callable, Dict, List, Optional, Sequence

import attr
from kubernetes import client as k8s_client
from kubernetes import config as k8s_config

from xmanager import xm
from xmanager.cloud import utils as cloud_utils
from xmanager.xm import utils
from xmanager.xm_local import executables as local_executables
from xmanager.xm_local import execution as local_execution
from xmanager.xm_local import executors as local_executors
from xmanager.xm_local import status as local_status


@functools.lru_cache()
def client():
  # Global singleton defers client creation until an actual launch.
  # If the user only launches local jobs, they don't need to create a client.
  return Client()


def _kubernetes_job_predicate(job: xm.Job) -> bool:
  return isinstance(job.executor, local_executors.Kubernetes)


def convert_to_valid_label(label: str) -> str:
  """Kubernetes labels must be RFC 1123 format compliant.

  https://github.com/kubernetes/kubernetes/blob/master/pkg/apis/apps/validation/validation.go
  A lowercase RFC 1123 label must consist of lower case alphanumeric characters
  or '-', and must start and end with an alphanumeric character (e.g. 'my-name',
  or '123-abc', regex used for validation is '[a-z0-9]([-a-z0-9]*[a-z0-9])?')

  Args:
    label: Label that may or may not be RFC 1123 compliant.

  Returns:
    A RFC 1123 format compliant label.
  """
  return label.replace('_', '-')


class Client:
  """Client class for interacting with Kubernetes."""

  def __init__(self, api_client: Optional[k8s_client.ApiClient] = None) -> None:
    if api_client is None:
      k8s_config.load_kube_config()
      api_client = k8s_client.ApiClient()
    self.api_client = api_client

  def launch(
      self,
      get_full_job_name: Callable[[str], str],
      jobs: Sequence[xm.Job],
  ) -> List[k8s_client.V1Job]:
    """Launches jobs on Kubernetes."""
    batch_jobs = []
    service = 'experiments'
    hostnames = [f'workerpool{i}' for i in range(len(jobs))]
    domains = [
        f'{host}.{service}.default.svc.cluster.local:2222' for host in hostnames
    ]
    specs = cloud_utils.create_cluster_specs(domains)
    for i, job in enumerate(jobs):
      executable = job.executable
      executor = job.executor
      if not isinstance(executable,
                        local_executables.GoogleContainerRegistryImage):
        raise ValueError('Executable {} has type {}. Executable must be of '
                         'type GoogleContainerRegistryImage.'.format(
                             executable, type(executable)))
      all_env_vars = {**executable.env_vars, **job.env_vars}
      env = [k8s_client.V1EnvVar(k, v) for k, v in all_env_vars.items()]
      env.append(k8s_client.V1EnvVar('CLUSTER_SPEC', specs[i]))
      job_name = convert_to_valid_label(get_full_job_name(job.name))
      container = k8s_client.V1Container(
          name=job_name,
          image=executable.image_path,
          resources=requirements_from_executor(executor),
          args=xm.merge_args(executable.args,
                             job.args).to_list(utils.ARG_ESCAPER),
          env=env)
      k8s_job = k8s_client.V1Job()
      k8s_job.metadata = k8s_client.V1ObjectMeta(name=job_name)
      k8s_job.spec = k8s_client.V1JobSpec(
          template=k8s_client.V1PodTemplateSpec(
              metadata=k8s_client.V1ObjectMeta(
                  labels={'service': service},
                  annotations=annotations_from_executor(executor),
              ),
              spec=k8s_client.V1PodSpec(
                  hostname=hostnames[i],
                  subdomain=service,
                  restart_policy='Never',
                  containers=[container],
                  node_selector=node_selector_from_executor(executor),
              ),
          ),
          backoff_limit=0,
      )
      batch_jobs.append(k8s_job)

    self._create_service(service)
    batch_api = k8s_client.BatchV1Api(self.api_client)
    for k8s_job in batch_jobs:
      batch_api.create_namespaced_job(namespace='default', body=k8s_job)
    return batch_jobs

  def _create_service(self, service: str) -> None:
    """Creates a K8s service with an `service: {service}` selector."""
    core_api = k8s_client.CoreV1Api(self.api_client)
    body = k8s_client.V1Service(
        metadata=k8s_client.V1ObjectMeta(name=service),
        spec=k8s_client.V1ServiceSpec(
            selector={'service': service},
            cluster_ip='None',
        ),
    )
    response = core_api.list_namespaced_service(namespace='default')
    for item in response.items:
      # service already exists
      if item.metadata.name == service:
        return
    core_api.create_namespaced_service(namespace='default', body=body)

  async def wait_for_job(self, job: k8s_client.V1Job) -> None:
    batch_api = k8s_client.BatchV1Api(self.api_client)
    backoff = 5  # seconds
    while True:
      await asyncio.sleep(backoff)
      response = batch_api.read_namespaced_job_status(
          namespace='default', name=job.metadata.name)
      if response.status.completion_time:
        return


@attr.s(auto_attribs=True)
class KubernetesHandle(local_execution.ExecutionHandle):
  """A handle for referring to the launched container."""

  jobs: List[k8s_client.V1Job]

  async def wait(self) -> None:
    await asyncio.gather(*[client().wait_for_job(job) for job in self.jobs])

  def get_status(self) -> local_status.LocalWorkUnitStatus:
    raise NotImplementedError


# Must act on all jobs with `local_executors.Kubernetes` executor.
def launch(get_full_job_name: Callable[[str], str],
           job_group: xm.JobGroup) -> List[KubernetesHandle]:
  """Launch K8s jobs in the job_group and return a handler."""
  jobs = xm.job_operators.collect_jobs_by_filter(job_group,
                                                 _kubernetes_job_predicate)
  # As client creation may throw, do not initiate it if there are no jobs.
  if not jobs:
    return []
  k8_jobs = client().launch(
      get_full_job_name=get_full_job_name,
      jobs=jobs,
  )
  return [KubernetesHandle(jobs=k8_jobs)]


def requirements_from_executor(
    executor: local_executors.Kubernetes) -> k8s_client.V1ResourceRequirements:
  """Get resource limits from the executor."""
  limits = {}
  for resource, value in executor.requirements.task_requirements.items():
    if resource in xm.GpuType:
      # TODO: Implement detection of whether an accelerator is an Nvidia
      # GPU. amd.com/gpu is another type of GPU that is not present in GCP.
      limits['nvidia.com/gpu'] = f'{value:g}'
    elif resource in xm.TpuType:
      pass
    else:
      # Converts resource amount to a string accepted by Kubernetes:
      # https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#meaning-of-cpu
      # https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#meaning-of-memory
      limits[str(resource).lower()] = f'{value:.15g}'

  return k8s_client.V1ResourceRequirements(limits=limits)


def annotations_from_executor(
    executor: local_executors.Kubernetes) -> Dict[str, str]:
  """Get Pod annotations from the executor for TPUs."""
  if executor.cloud_provider != local_executors.GOOGLE_KUBERNETES_ENGINE_CLOUD_PROVIDER:
    return {}

  if executor.requirements.accelerator in xm.TpuType:
    tpu_runtime_version = 'nightly'
    if executor.tpu_capability:
      tpu_runtime_version = executor.tpu_capability.tpu_runtime_version
    return {'tf-version.cloud-tpus.google.com': tpu_runtime_version}
  return {}


def node_selector_from_executor(
    executor: local_executors.Kubernetes) -> Dict[str, str]:
  """Get Pod annotations from the executor for TPUs."""
  if executor.cloud_provider != local_executors.GOOGLE_KUBERNETES_ENGINE_CLOUD_PROVIDER:
    return {}

  for resource in executor.requirements.task_requirements:
    if resource in xm.GpuType:
      return {
          'cloud.google.com/gke-accelerator':
              'nvidia-tesla-' + str(resource).lower()
      }
  return {}
